from __future__ import annotations

import os
import subprocess
import sys
import textwrap
from pathlib import Path


def test_libyara_modules_report_missing_backend_in_subprocess(tmp_path: Path) -> None:
    blocker = tmp_path / "yara.py"
    blocker.write_text('raise ImportError("blocked yara")\n', encoding="utf-8")

    code = textwrap.dedent(
        """
        import os
        import sys

        sys.path.insert(0, os.getcwd())

        import yaraast.libyara as libyara
        from yaraast.libyara.compiler import LibyaraCompiler
        from yaraast.libyara.direct_compiler import DirectASTCompiler, OptimizedMatcher
        from yaraast.libyara.scanner import LibyaraScanner

        print(f"available={libyara.YARA_AVAILABLE}")

        try:
            libyara.EquivalenceTester
        except Exception as exc:
            print(f"attr={type(exc).__name__}:{exc}")

        for cls in [LibyaraCompiler, DirectASTCompiler, LibyaraScanner]:
            try:
                cls()
            except Exception as exc:
                print(f"{cls.__name__}={type(exc).__name__}:{exc}")

        try:
            OptimizedMatcher(object())
        except Exception as exc:
            print(f"OptimizedMatcher={type(exc).__name__}:{exc}")
        """,
    )

    env = os.environ.copy()
    env["PYTHONPATH"] = (
        str(tmp_path) + os.pathsep + os.getcwd() + os.pathsep + env.get("PYTHONPATH", "")
    )

    result = subprocess.run(
        [sys.executable, "-c", code],
        cwd=os.getcwd(),
        env=env,
        capture_output=True,
        text=True,
        check=True,
    )

    stdout = result.stdout
    assert "available=False" in stdout
    assert "attr=ImportError:'EquivalenceTester' requires yara-python." in stdout
    assert "LibyaraCompiler=ImportError:yara-python is not installed." in stdout
    assert "DirectASTCompiler=ImportError:yara-python is not installed." in stdout
    assert "LibyaraScanner=ImportError:yara-python is not installed." in stdout
    assert "OptimizedMatcher=ImportError:yara-python is not installed." in stdout



"""Condition diff helpers."""

from __future__ import annotations

__all__ = ["condition_hashes", "emit_condition_diff"]


def condition_hashes(old_rule, new_rule, hasher) -> tuple[str, str]:
    """Hash conditions for comparison."""
    old_condition_hash = hasher.visit(old_rule.condition) if old_rule.condition else ""
    new_condition_hash = hasher.visit(new_rule.condition) if new_rule.condition else ""
    return old_condition_hash, new_condition_hash


def emit_condition_diff(
    base_path: str,
    result,
    diff_node,
    diff_type,
    old_condition_hash: str,
    new_condition_hash: str,
) -> None:
    """Record condition diff."""
    result.differences.append(
        diff_node(
            path=f"{base_path}/condition",
            diff_type=diff_type.MODIFIED,
            old_value=old_condition_hash,
            new_value=new_condition_hash,
            node_type="RuleCondition",
        ),
    )



"""More tests for performance string analyzer (no mocks)."""

from __future__ import annotations

from textwrap import dedent

from yaraast.parser import Parser
from yaraast.performance.string_analyzer import StringPatternAnalyzer, analyze_rule_performance


def _parse_yara(code: str):
    parser = Parser()
    return parser.parse(dedent(code))


def test_string_pattern_analyzer_patterns_and_stats() -> None:
    analyzer = StringPatternAnalyzer()
    patterns = [
        "alpha",
        "alpha",
        "alphabet",
        "beta",
        "zeta",
        "delta",
        "alpha_suffix",
        "beta_suffix",
    ]
    result = analyzer.analyze_patterns(patterns)

    assert result["total"] == len(patterns)
    assert result["duplicates"]["alpha"] == 2
    assert result["common_prefixes"]
    assert result["common_suffixes"]
    assert result["length_statistics"]["min"] > 0

    stats = analyzer.get_statistics()
    assert stats["total_strings"] == len(patterns)
    analyzer.reset_statistics()
    assert analyzer.get_statistics()["total_strings"] == 0


def test_string_pattern_analyzer_rule_and_file() -> None:
    code = """
    rule perf_rules {
        strings:
            $a = "aa"
            $b = "bbb"
            $c = /ab+c/
            $d = { 6A 40 ?? 0F }
        condition:
            any of them
    }
    rule perf_rules2 {
        strings:
            $a = "aa"
        condition:
            $a
    }
    """
    ast = _parse_yara(code)
    analyzer = StringPatternAnalyzer()

    rule_analysis = analyzer.analyze_rule(ast.rules[0])
    assert rule_analysis["rule"] == "perf_rules"
    assert rule_analysis["pattern_types"]["plain"] >= 1

    file_analysis = analyzer.analyze_file(ast)
    assert file_analysis["global"]["total"] >= 1
    assert file_analysis["cross_rule"]["total_shared"] >= 1


def test_analyze_rule_performance_issues() -> None:
    code = """
    rule perf_issue {
        strings:
            $a = "aa"
            $b = /ab+c/
        condition:
            $a or $b
    }
    """
    ast = _parse_yara(code)
    issues = analyze_rule_performance(ast.rules[0])

    assert any(issue.issue_type == "short_string" for issue in issues)
    assert any(issue.issue_type == "expensive_regex" for issue in issues)



"""Services for performance-check CLI (logic without IO)."""

from __future__ import annotations

from enum import StrEnum
from pathlib import Path
from typing import Any

from yaraast.cli.utils import read_text
from yaraast.parser.error_tolerant_parser import ErrorTolerantParser
from yaraast.performance.string_analyzer import StringPerformanceIssue, analyze_rule_performance


class Severity(StrEnum):
    """Issue severity levels."""

    WARNING = "warning"
    CRITICAL = "critical"
    INFO = "info"


def parse_performance_file(input_file: Path) -> Any:
    """Parse a YARA file and return AST or None."""
    content = read_text(input_file)
    parser = ErrorTolerantParser()
    ast, _, _ = parser.parse_with_errors(content)
    return ast


def analyze_rule_issues(rule: Any) -> list[StringPerformanceIssue]:
    """Analyze a single rule for performance issues."""
    return analyze_rule_performance(rule)


def filter_issues(
    issues: list[StringPerformanceIssue],
    severity: str | Severity,
    limit: int | None,
) -> list[StringPerformanceIssue]:
    """Filter issues by severity and limit.

    Accepts both Severity enum values and plain strings for backward
    compatibility (Severity inherits from str).
    """
    if severity == Severity.WARNING:
        issues = [i for i in issues if i.severity == Severity.WARNING]
    elif severity == Severity.CRITICAL:
        issues = [i for i in issues if i.severity == Severity.CRITICAL]

    if limit:
        issues = issues[:limit]

    return issues


def summarize_issues(issues: list[StringPerformanceIssue]) -> dict[str, Any]:
    """Summarize issues by type and severity."""
    issue_types: dict[str, dict[str, Any]] = {}
    for issue in issues:
        if issue.issue_type not in issue_types:
            issue_types[issue.issue_type] = {
                "count": 0,
                "critical": 0,
                "rules": set(),
            }
        entry = issue_types[issue.issue_type]
        entry["count"] += 1
        if issue.severity == "critical":
            entry["critical"] += 1
        entry["rules"].add(issue.rule_name)

    return issue_types



"""Semantic tokens provider for YARA Language Server."""

from __future__ import annotations

import logging
import time

from lsprotocol.types import Range, SemanticTokens, SemanticTokensLegend

from yaraast.lexer.lexer import Lexer
from yaraast.lexer.tokens import TokenType
from yaraast.lsp.runtime import LspRuntime
from yaraast.lsp.semantic_tokens_helpers import (
    encode_tokens,
    encode_tokens_in_range,
    map_token_type,
)

logger = logging.getLogger(__name__)

# Define semantic token types and modifiers according to LSP spec
TOKEN_TYPES = [
    "namespace",  # 0
    "class",  # 1 - rules
    "enum",  # 2
    "interface",  # 3
    "struct",  # 4
    "typeParameter",  # 5
    "type",  # 6
    "parameter",  # 7
    "variable",  # 8 - string identifiers
    "property",  # 9 - meta keys
    "enumMember",  # 10
    "decorator",  # 11
    "event",  # 12
    "function",  # 13 - builtin functions
    "method",  # 14
    "macro",  # 15
    "label",  # 16
    "comment",  # 17
    "string",  # 18
    "keyword",  # 19
    "number",  # 20
    "regexp",  # 21
    "operator",  # 22
]

TOKEN_MODIFIERS = [
    "declaration",  # 0
    "definition",  # 1
    "readonly",  # 2
    "static",  # 3
    "deprecated",  # 4
    "abstract",  # 5
    "async",  # 6
    "modification",  # 7
    "documentation",  # 8
    "defaultLibrary",  # 9
]


class SemanticTokensProvider:
    """Provides semantic tokens for advanced syntax highlighting."""

    def __init__(self, runtime: LspRuntime | None = None) -> None:
        self.runtime = runtime

    @staticmethod
    def get_legend() -> SemanticTokensLegend:
        """Get the semantic tokens legend."""
        return SemanticTokensLegend(
            token_types=TOKEN_TYPES,
            token_modifiers=TOKEN_MODIFIERS,
        )

    def get_semantic_tokens(self, text: str, uri: str | None = None) -> SemanticTokens:
        """
        Get semantic tokens for the document.

        Args:
            text: The YARA source code

        Returns:
            SemanticTokens with token information
        """
        ctx = self.runtime.ensure_document(uri, text) if self.runtime and uri else None
        if ctx is not None:
            cached = ctx.get_cached("semantic_tokens:full")
            if cached is not None:
                return cached

        tokens_data = []
        started = time.perf_counter()

        try:
            lexer = Lexer(text)
            tokens = lexer.tokenize()
            tokens_data = encode_tokens(tokens, self._map_token_type, TOKEN_TYPES)

        except Exception:
            logger.debug("Operation failed in %s", __name__, exc_info=True)
            # If tokenization fails, return empty tokens
            pass

        result = SemanticTokens(data=tokens_data)
        if ctx is not None:
            ctx.set_cached("semantic_tokens:full", result)
        if self.runtime is not None:
            self.runtime.record_latency(
                "semantic_tokens_full", (time.perf_counter() - started) * 1000.0
            )
        return result

    def get_semantic_tokens_range(
        self, text: str, range_: Range, uri: str | None = None
    ) -> SemanticTokens:
        """Get semantic tokens for a specific range."""
        ctx = self.runtime.ensure_document(uri, text) if self.runtime and uri else None
        cache_key = None
        if ctx is not None:
            cache_key = (
                f"semantic_tokens:range:{range_.start.line}:{range_.start.character}:"
                f"{range_.end.line}:{range_.end.character}"
            )
            cached = ctx.get_cached(cache_key)
            if cached is not None:
                return cached

        tokens_data = []
        started = time.perf_counter()

        try:
            lexer = Lexer(text)
            tokens = lexer.tokenize()
            tokens_data = encode_tokens_in_range(tokens, range_, self._map_token_type, TOKEN_TYPES)

        except Exception:
            logger.debug("Operation failed in %s", __name__, exc_info=True)

        result = SemanticTokens(data=tokens_data)
        if ctx is not None and cache_key is not None:
            ctx.set_cached(cache_key, result)
        if self.runtime is not None:
            self.runtime.record_latency(
                "semantic_tokens_range", (time.perf_counter() - started) * 1000.0
            )
        return result

    def _map_token_type(self, token_type: TokenType) -> str | None:
        return map_token_type(token_type)



"""AST visualization helpers for CLI output."""

from __future__ import annotations

from typing import Any

from rich.console import Console
from rich.tree import Tree

from yaraast.ast.base import YaraFile
from yaraast.cli.utils import format_json
from yaraast.cli.visitors import ASTDumper
from yaraast.errors import ValidationError


def print_ast(ast: YaraFile, console: Console | None = None) -> None:
    """Print AST in a readable format."""
    if console is None:
        console = Console()

    tree = Tree("YaraFile")
    _add_imports_to_tree(tree, ast.imports)
    _add_includes_to_tree(tree, ast.includes)
    _add_rules_to_tree(tree, ast.rules)
    console.print(tree)


def visualize_ast(ast: YaraFile, output_format: str = "json") -> str:
    """Visualize AST in a simple serializable format."""
    dumper = ASTDumper()
    data = dumper.visit(ast)

    if output_format == "json":
        return format_json(data, sort_keys=True)
    if output_format == "dict":
        return format_json(data, indent=None)

    msg = f"Unsupported output format: {output_format}"
    raise ValidationError(msg)


def _add_imports_to_tree(tree: Tree, imports: Any) -> None:
    """Add imports to the tree."""
    if not imports:
        return

    imports_branch = tree.add("imports")
    for imp in imports:
        imp_text = f'import "{imp.module}"'
        if hasattr(imp, "alias") and imp.alias:
            imp_text += f" as {imp.alias}"
        imports_branch.add(imp_text)


def _add_includes_to_tree(tree: Tree, includes: Any) -> None:
    """Add includes to the tree."""
    if not includes:
        return

    includes_branch = tree.add("includes")
    for inc in includes:
        includes_branch.add(f'include "{inc.path}"')


def _add_rules_to_tree(tree: Tree, rules: Any) -> None:
    """Add rules to the tree."""
    if not rules:
        return

    rules_branch = tree.add("rules")
    for rule in rules:
        rule_branch = _create_rule_branch(rules_branch, rule)
        _add_rule_components(rule_branch, rule)


def _create_rule_branch(rules_branch: Tree, rule: Any) -> Tree:
    """Create a rule branch in the tree."""
    rule_text = f"rule {rule.name}"
    if hasattr(rule, "modifiers") and rule.modifiers:
        rule_text = f"{' '.join(str(m) for m in rule.modifiers)} {rule_text}"
    return rules_branch.add(rule_text)


def _add_rule_components(rule_branch: Tree, rule: Any) -> None:
    """Add rule components to the rule branch."""
    _add_tags_to_rule(rule_branch, rule)
    _add_meta_to_rule(rule_branch, rule)
    _add_strings_to_rule(rule_branch, rule)
    _add_condition_to_rule(rule_branch, rule)


def _add_tags_to_rule(rule_branch: Tree, rule: Any) -> None:
    """Add tags to the rule branch."""
    if not (hasattr(rule, "tags") and rule.tags):
        return

    tags_branch = rule_branch.add("tags")
    for tag in rule.tags:
        tag_name = tag.name if hasattr(tag, "name") else str(tag)
        tags_branch.add(tag_name)


def _add_meta_to_rule(rule_branch: Tree, rule: Any) -> None:
    """Add meta to the rule branch."""
    if not (hasattr(rule, "meta") and rule.meta):
        return

    meta_branch = rule_branch.add("meta")
    for meta_item in rule.meta:
        if hasattr(meta_item, "key"):
            meta_branch.add(f"{meta_item.key} = {meta_item.value}")


def _add_strings_to_rule(rule_branch: Tree, rule: Any) -> None:
    """Add strings to the rule branch."""
    if not (hasattr(rule, "strings") and rule.strings):
        return

    strings_branch = rule_branch.add("strings")
    for string_def in rule.strings:
        string_type = type(string_def).__name__
        strings_branch.add(f"{string_def.identifier} ({string_type})")


def _add_condition_to_rule(rule_branch: Tree, rule: Any) -> None:
    """Add condition to the rule branch."""
    if not (hasattr(rule, "condition") and rule.condition):
        return

    condition_branch = rule_branch.add("condition")
    condition_branch.add(str(rule.condition))


__all__ = [
    "_add_condition_to_rule",
    "_add_imports_to_tree",
    "_add_includes_to_tree",
    "_add_meta_to_rule",
    "_add_rule_components",
    "_add_rules_to_tree",
    "_add_strings_to_rule",
    "_add_tags_to_rule",
    "_create_rule_branch",
    "print_ast",
    "visualize_ast",
]



"""Additional real coverage for unified_parser and semantic validator core."""

from __future__ import annotations

import sys
from pathlib import Path

import pytest

from yaraast.ast.base import Location
from yaraast.types.semantic_validator_core import ValidationError, ValidationResult
from yaraast.unified_parser import UnifiedParser


def test_validation_error_without_location_and_combine_invalid_result() -> None:
    err = ValidationError("broken")
    assert str(err) == "error: broken"

    left = ValidationResult()
    right = ValidationResult()
    right.add_error("x")
    left.combine(right)
    assert left.is_valid is False
    assert left.total_issues == 1


def test_validation_error_to_dict_with_location_and_suggestion() -> None:
    err = ValidationError(
        "broken",
        location=Location(line=2, column=3, file="sample.yar"),
        suggestion="fix it",
    )
    data = err.to_dict()
    assert data["location"] == {"file": "sample.yar", "line": 2, "column": 3}
    assert data["suggestion"] == "fix it"


@pytest.mark.skipif(sys.platform == "win32", reason="chmod not effective on Windows")
def test_unified_parser_parse_file_permission_and_oserror(tmp_path: Path) -> None:
    restricted_dir = tmp_path / "restricted"
    restricted_dir.mkdir()
    restricted_file = restricted_dir / "r.yar"

    restricted_dir.chmod(0)
    try:
        with pytest.raises(PermissionError, match="Permission denied reading file"):
            UnifiedParser.parse_file(restricted_file)
    finally:
        restricted_dir.chmod(0o755)

    too_long_name = "a" * 5000
    with pytest.raises(OSError, match="Error accessing file"):
        UnifiedParser.parse_file(tmp_path / too_long_name)


def test_unified_parser_parse_yaral_branch() -> None:
    source = """
    rule demo {
        events:
            $e.metadata.event_type = "LOGIN"
        condition:
            #e > 1
    }
    """
    ast = UnifiedParser(source).parse()
    assert len(ast.rules) == 1



"""Additional tests for YARA-L validator and optimizer."""

from __future__ import annotations

from yaraast.yaral.ast_nodes import (
    BinaryCondition,
    ConditionSection,
    EventAssignment,
    EventExistsCondition,
    EventsSection,
    EventVariable,
    MatchSection,
    MatchVariable,
    MetaEntry,
    MetaSection,
    OptionsSection,
    OutcomeAssignment,
    OutcomeSection,
    TimeWindow,
    UDMFieldPath,
    UnaryCondition,
    YaraLFile,
    YaraLRule,
)
from yaraast.yaral.optimizer import YaraLOptimizer
from yaraast.yaral.validator import YaraLValidator


def test_yaral_validator_flags_multiple_issues() -> None:
    event_var = EventVariable(name="$e")
    bad_assignment = EventAssignment(
        event_var=event_var,
        field_path=UDMFieldPath(parts=["unknown", "field"]),
        operator="==",
        value="LOGIN",
    )
    regex_warning_assignment = EventAssignment(
        event_var=event_var,
        field_path=UDMFieldPath(parts=["metadata", "event_type"]),
        operator="=~",
        value="LOGIN",
    )

    rule1 = YaraLRule(
        name="1bad",
        meta=MetaSection(entries=[MetaEntry(key="severity", value="urgent")]),
        events=EventsSection(statements=[bad_assignment, regex_warning_assignment]),
        match=MatchSection(
            variables=[MatchVariable(variable="user", time_window=TimeWindow(-1, "x"))]
        ),
        condition=ConditionSection(expression=EventExistsCondition(event="e")),
        outcome=OutcomeSection(
            assignments=[
                OutcomeAssignment(variable="score", expression=1),
                OutcomeAssignment(variable="score", expression=2),
            ]
        ),
        options=OptionsSection(options={"unknown_option": True}),
    )

    rule2 = YaraLRule(
        name="1bad",
        events=EventsSection(statements=[bad_assignment]),
        condition=ConditionSection(expression=EventExistsCondition(event="missing")),
    )

    errors, warnings = YaraLValidator().validate(YaraLFile(rules=[rule1, rule2]))

    error_messages = [err.message for err in errors]
    warning_messages = [warn.message for warn in warnings]

    assert any("Duplicate rule name" in msg for msg in error_messages)
    assert any(
        "Rule name '1bad' must start with letter or underscore" in msg for msg in error_messages
    )
    assert any("Invalid operator" in msg for msg in error_messages)
    assert any("Invalid time unit" in msg for msg in error_messages)
    assert any("Time window duration must be positive" in msg for msg in error_messages)
    assert any("Unknown UDM namespace" in msg for msg in warning_messages)
    assert any("Regex operator" in msg for msg in warning_messages)
    assert any("Unknown option" in msg for msg in warning_messages)
    assert any("Missing recommended meta field" in msg for msg in warning_messages)


def test_yaral_optimizer_simplifies_conditions_and_time_window() -> None:
    optimizer = YaraLOptimizer()

    condition = ConditionSection(
        expression=BinaryCondition(
            operator="and",
            left=EventExistsCondition(event="e"),
            right=EventExistsCondition(event="e"),
        )
    )
    double_negation = ConditionSection(
        expression=UnaryCondition(
            operator="not",
            operand=UnaryCondition(operator="not", operand=EventExistsCondition(event="e")),
        )
    )

    match = MatchSection(
        variables=[MatchVariable(variable="user", time_window=TimeWindow(7200, "s"))]
    )

    rule = YaraLRule(
        name="opt",
        events=EventsSection(statements=[]),
        match=match,
        condition=condition,
        outcome=OutcomeSection(assignments=[]),
    )

    optimized, stats = optimizer.optimize(YaraLFile(rules=[rule]))

    assert stats.conditions_simplified >= 1
    assert stats.time_windows_optimized >= 1
    optimized_match = optimized.rules[0].match
    assert optimized_match is not None
    assert optimized_match.variables[0].time_window.unit == "h"

    # Verify double negation path via direct call
    optimized_double = optimizer._optimize_condition_section(double_negation)
    assert isinstance(optimized_double.expression, EventExistsCondition)



"""Reporting helpers for semantic CLI output."""

from __future__ import annotations

from pathlib import Path
from typing import Any

import click

from yaraast.cli.utils import format_json, write_text


def display_validation_start(file_path: Path, quiet: bool) -> None:
    if not quiet:
        click.echo(f"Validating {file_path}...")


def display_parse_failure(file_path: Path) -> None:
    click.echo(f"Error: Failed to parse {file_path}", err=True)


def display_processing_error(file_path: Path, error: Exception) -> None:
    click.echo(f"Error processing {file_path}: {error}", err=True)


def display_text_results(
    file_path: Path,
    result: Any,
    show_warnings: bool,
    show_suggestions: bool,
    quiet: bool,
) -> None:
    if result.errors:
        for error in result.errors:
            click.echo(click.style(str(error), fg="red"), err=True)
            if show_suggestions and error.suggestion:
                click.echo(click.style(f"  Suggestion: {error.suggestion}", fg="blue"))

    if show_warnings and result.warnings:
        for warning in result.warnings:
            click.echo(click.style(str(warning), fg="yellow"))
            if show_suggestions and warning.suggestion:
                click.echo(
                    click.style(f"  Suggestion: {warning.suggestion}", fg="blue"),
                )

    if not quiet:
        if result.is_valid and not result.warnings:
            click.echo(click.style(f"✓ {file_path}: All checks passed", fg="green"))
        elif result.is_valid:
            click.echo(
                click.style(
                    f"✓ {file_path}: Valid with {len(result.warnings)} warnings",
                    fg="yellow",
                ),
            )
        else:
            click.echo(
                click.style(f"✗ {file_path}: {len(result.errors)} errors", fg="red"),
            )


def display_summary(total_files: int, total_errors: int, total_warnings: int) -> None:
    click.echo()
    click.echo(f"Validated {total_files} file(s)")

    if total_errors > 0:
        click.echo(click.style(f"Found {total_errors} errors", fg="red"))

    if total_warnings > 0:
        click.echo(click.style(f"Found {total_warnings} warnings", fg="yellow"))

    if total_errors == 0 and total_warnings == 0:
        click.echo(click.style("All files passed validation", fg="green"))


def write_output_file(output_path: Path, results: list[dict], format: str) -> None:
    if format == "json":
        write_text(output_path, format_json(results))
        return

    lines: list[str] = []
    for result in results:
        lines.append(f"File: {result['file']}")
        lines.append(f"Valid: {result['is_valid']}")
        lines.append(f"Errors: {len(result['errors'])}")
        lines.append(f"Warnings: {len(result['warnings'])}")

        for error in result["errors"]:
            lines.append(f"ERROR: {error['message']}")
            if error.get("location"):
                loc = error["location"]
                lines.append(
                    "  Location: "
                    f"{loc.get('file', 'unknown')}:{loc.get('line', 0)}:{loc.get('column', 0)}",
                )
            if error.get("suggestion"):
                lines.append(f"  Suggestion: {error['suggestion']}")

        for warning in result["warnings"]:
            lines.append(f"WARNING: {warning['message']}")
            if warning.get("location"):
                loc = warning["location"]
                lines.append(
                    "  Location: "
                    f"{loc.get('file', 'unknown')}:{loc.get('line', 0)}:{loc.get('column', 0)}",
                )
            if warning.get("suggestion"):
                lines.append(f"  Suggestion: {warning['suggestion']}")

        lines.append("")

    write_text(output_path, "\n".join(lines))


def emit_json_results(results: list[dict[str, Any]]) -> None:
    click.echo(format_json(results))



"""Additional tests for pragma nodes (no mocks)."""

from __future__ import annotations

from yaraast.ast.pragmas import (
    ConditionalDirective,
    CustomPragma,
    DefineDirective,
    IncludeOncePragma,
    InRulePragma,
    Pragma,
    PragmaBlock,
    PragmaScope,
    PragmaType,
    UndefDirective,
    create_define,
    create_endif,
    create_ifdef,
    create_ifndef,
    create_in_rule_pragma,
    create_include_once,
    create_pragma,
    create_undef,
)


def test_pragma_string_reprs_and_flags() -> None:
    assert PragmaType.from_string("not_known") == PragmaType.CUSTOM

    generic = Pragma(PragmaType.UNDEF, "undef", ["X"])
    assert str(generic) == "#undef X"

    pragma_style = Pragma(PragmaType.PRAGMA, "optimize", ["fast"])
    assert str(pragma_style) == "#pragma optimize fast"

    include_once = IncludeOncePragma()
    assert include_once.is_include_once is True
    assert str(include_once) == "#include_once"

    define = DefineDirective("X", "1")
    assert define.is_define is True
    assert str(define) == "#define X 1"
    assert str(DefineDirective("ONLY_NAME")) == "#define ONLY_NAME"

    undef = UndefDirective("X")
    assert str(undef) == "#undef X"

    ifdef = ConditionalDirective.ifdef("FOO")
    ifndef = ConditionalDirective.ifndef("BAR")
    endif = ConditionalDirective.endif()
    assert str(ifdef) == "#ifdef FOO"
    assert str(ifndef) == "#ifndef BAR"
    assert str(endif) == "#endif"


def test_custom_pragma_and_block_helpers() -> None:
    custom = CustomPragma(name="vendor", arguments=["x", "y"], scope=PragmaScope.FILE)
    custom.set_parameter("level", 3)
    assert custom.get_parameter("level") == 3
    assert str(custom).startswith("#pragma vendor")

    block = PragmaBlock(scope=PragmaScope.RULE)
    block.add_pragma(custom)
    assert block.has_pragma(PragmaType.CUSTOM) is True
    assert block.get_pragmas_by_type(PragmaType.CUSTOM) == [custom]
    assert block.get_pragmas_by_type(PragmaType.DEFINE) == []
    assert str(block) == str(custom)


def test_create_helpers_and_in_rule_positions() -> None:
    pragma = create_pragma("define", ["X", "2"])
    assert isinstance(pragma, Pragma)
    assert pragma.pragma_type == PragmaType.DEFINE

    include_once = create_include_once()
    define = create_define("FLAG")
    undef = create_undef("OLD")
    ifdef = create_ifdef("FEATURE")
    ifndef = create_ifndef("NO_FEATURE")
    endif = create_endif()
    assert isinstance(include_once, IncludeOncePragma)
    assert isinstance(define, DefineDirective)
    assert isinstance(undef, UndefDirective)
    assert isinstance(ifdef, ConditionalDirective)
    assert isinstance(ifndef, ConditionalDirective)
    assert isinstance(endif, ConditionalDirective)

    custom = create_pragma("vendor_specific", ["x"], scope=PragmaScope.LOCAL)
    assert isinstance(custom, CustomPragma)
    assert custom.scope == PragmaScope.LOCAL

    in_rule = InRulePragma(pragma=pragma, position="before_strings")
    assert in_rule.is_before_strings is True
    assert in_rule.is_after_strings is False
    assert in_rule.is_before_condition is False
    assert str(in_rule) == str(pragma)

    in_rule2 = create_in_rule_pragma(pragma, "before_condition")
    assert in_rule2.is_before_condition is True


def test_pragma_accept_methods() -> None:
    class _Visitor:
        def visit_pragma(self, node):
            return ("pragma", node.name)

        def visit_in_rule_pragma(self, node):
            return ("in_rule", node.position)

        def visit_pragma_block(self, node):
            return ("block", len(node.pragmas))

    visitor = _Visitor()
    pragma = Pragma(PragmaType.DEFINE, "define", ["X"])
    in_rule = InRulePragma(pragma=pragma, position="after_strings")
    block = PragmaBlock(pragmas=[pragma])

    assert pragma.accept(visitor) == ("pragma", "define")
    assert in_rule.accept(visitor) == ("in_rule", "after_strings")
    assert block.accept(visitor) == ("block", 1)



from __future__ import annotations

import pytest

from yaraast.libyara.compiler import YARA_AVAILABLE, LibyaraCompiler


@pytest.mark.skipif(not YARA_AVAILABLE, reason="yara-python not available")
def test_libyara_compiler_handles_read_and_save_failures(tmp_path) -> None:
    compiler = LibyaraCompiler()

    directory = tmp_path / "adir"
    directory.mkdir()
    read_result = compiler.compile_file(directory)
    assert read_result.success is False
    assert read_result.errors
    assert "Error reading file" in read_result.errors[0]

    assert compiler.save_compiled_rules(object(), tmp_path / "rules.bin") is False


@pytest.mark.skipif(not YARA_AVAILABLE, reason="yara-python not available")
def test_libyara_compiler_handles_type_error_and_null_byte_source_paths() -> None:
    compiler = LibyaraCompiler()

    type_error_result = compiler.compile_source(None)  # type: ignore[arg-type]
    assert type_error_result.success is False
    assert type_error_result.errors
    assert "Unexpected error" in type_error_result.errors[0]

    null_byte_result = compiler.compile_source(
        'rule bad {\nstrings:\n$a = "a\x00b"\ncondition:\n$a\n}\n'
    )
    assert null_byte_result.source_code is not None



"""Helper utilities for YARA completion providers."""

from __future__ import annotations

import logging
from typing import Any

from lsprotocol.types import CompletionItem, CompletionItemKind, InsertTextFormat, Position

from yaraast.errors import ParseError
from yaraast.lsp.language_services import parse_source
from yaraast.lsp.lsp_docs import BUILTIN_DOCS, KEYWORD_DOCS, MODULE_DOCS
from yaraast.lsp.runtime import LanguageMode

logger = logging.getLogger(__name__)

try:
    from yaraast.ast.conditions import ForExpression
except Exception:  # pragma: no cover
    ForExpression = None  # type: ignore[assignment,misc]

try:
    from yaraast.yarax.ast_nodes import WithStatement
except Exception:  # pragma: no cover
    WithStatement = None  # type: ignore[assignment,misc]

KEYWORDS: list[str] = list(KEYWORD_DOCS.keys())
BUILTIN_FUNCTIONS: dict[str, str] = dict(BUILTIN_DOCS)
YARAL_KEYWORDS: list[str] = [
    "rule",
    "meta",
    "events",
    "match",
    "condition",
    "outcome",
    "options",
    "and",
    "or",
    "not",
]
YARAX_KEYWORDS: list[str] = list(dict.fromkeys([*KEYWORDS, "with", "lambda", "match"]))

STRING_MODIFIERS: dict[str, str] = {
    "nocase": "Case-insensitive string matching",
    "wide": "Match wide-character (UTF-16) strings",
    "ascii": "Match ASCII strings",
    "xor": "Match XOR-encoded strings",
    "base64": "Match base64-encoded strings",
    "base64wide": "Match base64-encoded wide strings",
    "fullword": "Match complete words only",
}

MODULE_COMPLETIONS: list[tuple[str, str]] = list(MODULE_DOCS.items())


def analyze_context(text: str, position: Position) -> str:
    """Analyze the context at the current position."""
    lines = text.split("\n")
    if position.line >= len(lines):
        return "general"

    current_line = lines[position.line]
    before_cursor = current_line[: position.character]

    if "import" in before_cursor and '"' in before_cursor:
        return "import"

    if "." in before_cursor:
        parts = before_cursor.rsplit(".", 1)
        if len(parts) == 2:
            return "module_member"

    if "$" in before_cursor and "=" in before_cursor:
        if any(mod in before_cursor for mod in ["nocase", "wide", "ascii", "xor"]):
            return "string_modifier"
        if '"' in before_cursor or "}}" in before_cursor or "/" in before_cursor:
            return "string_modifier"

    for i in range(position.line, -1, -1):
        if i < len(lines):
            line = lines[i].strip()
            if line.startswith("condition:"):
                return "condition"
            if line.startswith("meta:"):
                return "meta"
            if line.startswith("strings:"):
                return "strings"
            if line.startswith("rule "):
                break

    return "general"


def get_current_module(text: str, position: Position) -> str | None:
    """Get the module name before the current position."""
    lines = text.split("\n")
    if position.line >= len(lines):
        return None

    current_line = lines[position.line]
    before_cursor = current_line[: position.character]

    if "." in before_cursor:
        parts = before_cursor.rsplit(".", 1)
        if len(parts) == 2:
            tokens = parts[0].split()
            if tokens:
                module_name = tokens[-1]
                return "".join(c for c in module_name if c.isalnum() or c == "_")

    return None


_SNIPPET_TEMPLATES: dict[str, tuple[str, str]] = {
    "rule": (
        'rule ${1:rule_name} {\n\tmeta:\n\t\t${2:author = "analyst"}\n\tstrings:\n\t\t${3:\\$s = "pattern"}\n\tcondition:\n\t\t${0:any of them}\n}',
        "Create a new YARA rule with meta, strings, and condition",
    ),
    "meta": ("meta:\n\t${0}", "Start a metadata section"),
    "strings": ("strings:\n\t${0}", "Start a strings definition section"),
    "condition": ("condition:\n\t${0:true}", "Start a condition section"),
}


def build_keyword_completions(keywords: list[str]) -> list[CompletionItem]:
    """Build keyword completions with snippet templates for structural keywords."""
    items: list[CompletionItem] = []
    for keyword in keywords:
        snippet = _SNIPPET_TEMPLATES.get(keyword)
        if snippet:
            insert_text, doc = snippet
            items.append(
                CompletionItem(
                    label=keyword,
                    kind=CompletionItemKind.Snippet,
                    detail="YARA template",
                    documentation=doc,
                    insert_text=insert_text,
                    insert_text_format=InsertTextFormat.Snippet,
                )
            )
        items.append(
            CompletionItem(
                label=keyword,
                kind=CompletionItemKind.Keyword,
                detail="YARA keyword",
                insert_text=keyword,
            )
        )
    return items


def build_builtin_function_completions(
    builtin_functions: dict[str, str],
) -> list[CompletionItem]:
    """Build built-in function completions."""
    return [
        CompletionItem(
            label=func_name,
            kind=CompletionItemKind.Function,
            detail="Built-in function",
            documentation=description,
            insert_text=f"{func_name}($0)",
            insert_text_format=InsertTextFormat.Snippet,
        )
        for func_name, description in builtin_functions.items()
    ]


def build_module_completions() -> list[CompletionItem]:
    """Build completions for available YARA modules."""
    return [
        CompletionItem(
            label=module_name,
            kind=CompletionItemKind.Module,
            detail="YARA module",
            documentation=description,
            insert_text=module_name,
        )
        for module_name, description in MODULE_COMPLETIONS
    ]


def build_module_member_completions(
    module_name: str,
    module_def: Any,
    access_chain: str = "",
) -> list[CompletionItem]:
    """Build completions for module members with deep type introspection.

    Supports nested access: pe.sections[0]. resolves to StructType fields.
    """
    from yaraast.types._registry_collections import ArrayType, DictionaryType, StructType

    items: list[CompletionItem] = []
    prefix = f"{module_name}.{access_chain}" if access_chain else module_name

    # If access_chain points to a nested type, resolve it
    if access_chain:
        resolved_type = _resolve_access_chain(module_def, access_chain)
        if resolved_type is not None:
            items.extend(_completions_for_type(prefix, resolved_type))
            return items

    for func_name, func_def in module_def.functions.items():
        params = ", ".join(p[0] for p in func_def.parameters)
        items.append(
            CompletionItem(
                label=func_name,
                kind=CompletionItemKind.Function,
                detail=f"{prefix}.{func_name}({params})",
                documentation=getattr(func_def, "description", None),
                insert_text=f"{func_name}($0)",
                insert_text_format=InsertTextFormat.Snippet,
            )
        )

    fields = getattr(module_def, "fields", None)
    if fields is None:
        fields = getattr(module_def, "attributes", {})
    if fields is None:
        fields = {}
    for field_name, field_def in fields.items():
        kind = CompletionItemKind.Field
        detail_type = str(field_def)
        # Indicate navigable types
        if isinstance(field_def, ArrayType):
            detail_type = f"array[{field_def.element_type}]"
            if isinstance(field_def.element_type, StructType):
                detail_type += " (indexable)"
        elif isinstance(field_def, StructType):
            detail_type = "struct"
            kind = CompletionItemKind.Struct
        elif isinstance(field_def, DictionaryType):
            detail_type = f"dict[{field_def.key_type}, {field_def.value_type}]"
        items.append(
            CompletionItem(
                label=field_name,
                kind=kind,
                detail=f"{prefix}.{field_name}: {detail_type}",
                documentation=getattr(field_def, "description", None),
                insert_text=field_name,
            )
        )

    return items


def _resolve_access_chain(module_def: Any, chain: str) -> Any:
    """Resolve a dotted access chain like 'sections[0]' against module type definitions."""
    from yaraast.types._registry_collections import ArrayType, StructType

    fields = getattr(module_def, "fields", None) or getattr(module_def, "attributes", {}) or {}
    parts = chain.replace("]", "").split(".")
    current_type = None

    for part in parts:
        if not part:
            continue
        # Handle array indexing: sections[0
        field_name = part.split("[")[0] if "[" in part else part

        if current_type is None:
            current_type = fields.get(field_name)
        elif isinstance(current_type, StructType):
            current_type = current_type.fields.get(field_name)
        else:
            return None

        if current_type is None:
            return None

        # If indexed, descend into element type
        if "[" in part and isinstance(current_type, ArrayType):
            current_type = current_type.element_type

    return current_type


def _completions_for_type(prefix: str, type_def: Any) -> list[CompletionItem]:
    """Generate completion items from a resolved type definition."""
    from yaraast.types._registry_collections import StructType

    items: list[CompletionItem] = []
    if isinstance(type_def, StructType):
        for field_name, field_type in type_def.fields.items():
            items.append(
                CompletionItem(
                    label=field_name,
                    kind=CompletionItemKind.Field,
                    detail=f"{prefix}.{field_name}: {field_type}",
                    insert_text=field_name,
                )
            )
    return items


def build_string_modifier_completions(
    string_modifiers: dict[str, str],
) -> list[CompletionItem]:
    """Build string modifier completions."""
    return [
        CompletionItem(
            label=modifier,
            kind=CompletionItemKind.Property,
            detail="String modifier",
            documentation=description,
            insert_text=modifier,
        )
        for modifier, description in string_modifiers.items()
    ]


def build_condition_completions(text: str, keywords: list[str]) -> list[CompletionItem]:
    """Build completions relevant to condition context."""
    items: list[CompletionItem] = []
    ast = None

    try:
        ast = parse_source(text)
        if not ast:
            raise ParseError("Failed to parse")
        items.extend(_string_identifier_completions(ast))
    except Exception:
        logger.debug("Operation failed in %s", __name__, exc_info=True)

    try:
        if ast:
            items.extend(_loop_variable_completions(ast))
    except Exception:
        logger.debug("Operation failed in %s", __name__, exc_info=True)

    items.extend(build_keyword_completions(keywords))
    return items


def _string_identifier_completions(ast: Any) -> list[CompletionItem]:
    """Build completions for string identifiers defined in the rule."""
    items: list[CompletionItem] = []
    for rule in ast.rules:
        for string_def in rule.strings:
            identifier = string_def.identifier
            items.append(
                CompletionItem(
                    label=identifier,
                    kind=CompletionItemKind.Variable,
                    detail="String identifier",
                    insert_text=identifier,
                )
            )
            if identifier.startswith("$"):
                base_name = identifier[1:]
                items.extend(
                    [
                        CompletionItem(
                            label=f"#{base_name}",
                            kind=CompletionItemKind.Variable,
                            detail="String count",
                            insert_text=f"#{base_name}",
                        ),
                        CompletionItem(
                            label=f"@{base_name}",
                            kind=CompletionItemKind.Variable,
                            detail="String offset",
                            insert_text=f"@{base_name}",
                        ),
                        CompletionItem(
                            label=f"!{base_name}",
                            kind=CompletionItemKind.Variable,
                            detail="String length",
                            insert_text=f"!{base_name}",
                        ),
                    ]
                )
    return items


def _loop_variable_completions(ast: Any) -> list[CompletionItem]:
    """Build completions for loop variables from for-expressions in conditions."""
    items: list[CompletionItem] = []
    for rule in ast.rules:
        if rule.condition is not None:
            for var_name in _extract_loop_variables(rule.condition):
                items.append(
                    CompletionItem(
                        label=var_name,
                        kind=CompletionItemKind.Variable,
                        detail="Loop variable",
                        insert_text=var_name,
                    )
                )
    return items


def _extract_loop_variables(node: Any) -> list[str]:
    """Recursively walk an AST condition and extract variables from for-loops and with-declarations."""
    variables: list[str] = []
    if node is None:
        return variables
    if ForExpression is not None and isinstance(node, ForExpression) and node.variable:
        variables.append(node.variable)
    if WithStatement is not None and isinstance(node, WithStatement):
        for decl in node.declarations:
            if hasattr(decl, "identifier") and decl.identifier:
                variables.append(decl.identifier)
    # Recurse into child nodes
    children_method = getattr(node, "children", None)
    if callable(children_method):
        try:
            for child in children_method():
                variables.extend(_extract_loop_variables(child))
        except Exception:
            logger.debug("Operation failed in %s", __name__, exc_info=True)
    return variables


def get_keywords_for_mode(language_mode: LanguageMode) -> list[str]:
    """Return keyword set appropriate for the active language mode."""
    if language_mode is LanguageMode.YARA_L:
        return YARAL_KEYWORDS
    if language_mode is LanguageMode.YARA_X:
        return YARAX_KEYWORDS
    return KEYWORDS



"""CLI command for parsing YARA files."""

from __future__ import annotations

import click
from rich.console import Console

from yaraast.cli.parse_reporting import (
    display_cli_error,
    generate_output_by_format,
    report_parsing_errors,
)
from yaraast.cli.parse_services import parse_content_by_dialect
from yaraast.cli.utils import read_text

console = Console()


@click.command()
@click.argument("input_file", type=click.Path(exists=True))
@click.option("-o", "--output", type=click.Path(), help="Output file (default: stdout)")
@click.option(
    "-f",
    "--format",
    "output_format",
    type=click.Choice(["yara", "json", "yaml", "tree"]),
    default="yara",
    help="Output format",
)
@click.option(
    "--dialect",
    type=click.Choice(["auto", "yara", "yara-x", "yara-l"]),
    default="auto",
    help="YARA dialect to use (auto-detect by default)",
)
def parse(input_file: str, output: str | None, output_format: str, dialect: str) -> None:
    """Parse a YARA file and output in various formats. Supports YARA, YARA-X, and YARA-L."""
    try:
        content = read_text(input_file)
        show_status = output_format not in {"json", "yaml"}
        ast, lexer_errors, parser_errors = parse_content_by_dialect(
            content,
            dialect,
            show_status,
            console.print,
        )
        if show_status:
            report_parsing_errors(lexer_errors, parser_errors, ast)
        if lexer_errors or parser_errors:
            raise SystemExit(1)
        generate_output_by_format(ast, output_format, output)

    except Exception as e:
        display_cli_error(console, e)
        raise click.Abort from None



"""Real integration tests for CLI lsp/serialize commands (no mocks)."""

from __future__ import annotations

from pathlib import Path

from click.testing import CliRunner

import yaraast.cli.commands.lsp as lsp_cmd
import yaraast.cli.commands.serialize as ser_cmd


def _rule(name: str, value: str = "abc") -> str:
    return f"""
rule {name} {{
    strings:
        $a = "{value}"
    condition:
        $a
}}
"""


def test_lsp_help_real() -> None:
    runner = CliRunner()
    result = runner.invoke(lsp_cmd.lsp, ["--help"])
    assert result.exit_code == 0
    assert "--stdio" in result.output
    assert "--tcp" in result.output


def test_serialize_diff_no_changes_real(tmp_path: Path) -> None:
    runner = CliRunner()
    old_file = tmp_path / "old.yar"
    new_file = tmp_path / "new.yar"
    old_file.write_text(_rule("same"), encoding="utf-8")
    new_file.write_text(_rule("same"), encoding="utf-8")

    result = runner.invoke(ser_cmd.serialize, ["diff", str(old_file), str(new_file)])
    assert result.exit_code == 0
    assert "No differences found" in result.output


def test_serialize_diff_patch_and_json_output_real(tmp_path: Path) -> None:
    runner = CliRunner()
    old_file = tmp_path / "old.yar"
    new_file = tmp_path / "new.yar"
    old_file.write_text(_rule("r1", "abc"), encoding="utf-8")
    new_file.write_text(_rule("r1", "xyz"), encoding="utf-8")

    patch_out = tmp_path / "changes.patch"
    result_patch = runner.invoke(
        ser_cmd.serialize,
        ["diff", str(old_file), str(new_file), "--patch", "-o", str(patch_out)],
    )
    assert result_patch.exit_code == 0
    assert patch_out.exists()
    assert patch_out.read_text(encoding="utf-8")

    json_out = tmp_path / "changes.json"
    result_json = runner.invoke(
        ser_cmd.serialize,
        ["diff", str(old_file), str(new_file), "-o", str(json_out), "-f", "json"],
    )
    assert result_json.exit_code == 0
    assert json_out.exists()
    assert json_out.read_text(encoding="utf-8")


def test_serialize_error_paths_real(tmp_path: Path) -> None:
    runner = CliRunner()
    bad = tmp_path / "bad.yar"
    bad.write_text("rule broken { condition:", encoding="utf-8")

    export_result = runner.invoke(ser_cmd.serialize, ["export", str(bad)])
    assert export_result.exit_code != 0

    validate_result = runner.invoke(ser_cmd.serialize, ["validate", str(bad)])
    assert validate_result.exit_code != 0

    info_result = runner.invoke(ser_cmd.serialize, ["info", str(bad)])
    assert info_result.exit_code != 0



"""Additional tests for Enhanced YARA-L parser paths."""

from __future__ import annotations

from textwrap import dedent

from yaraast.yaral.ast_nodes import ReferenceList, RegexPattern
from yaraast.yaral.enhanced_parser import EnhancedYaraLParser


def test_enhanced_parser_parses_events_conditions_outcome_options() -> None:
    yaral_code = """
    rule enhanced_extras {
        meta:
            author = "sec"
            enabled = false
            threshold = 3

        events:
            $e.metadata.event_type matches /LOGIN/i and $e.principal.user = "alice"
            $e.security_result.action = true
            $e.metadata.product_name = %products%

        condition:
            not ($e and #e > 1) or src.ip in %suspicious_ips%

        outcome:
            $count = count(metadata.event_type)
            if #e > 0 then "high" else "low"

        options:
            case_sensitive = false
            max_events = 10
            output_format = json
    }
    """

    parser = EnhancedYaraLParser(dedent(yaral_code))
    ast = parser.parse()

    assert parser.errors == []
    assert len(ast.rules) == 1
    rule = ast.rules[0]
    assert rule.name == "enhanced_extras"
    assert rule.meta is not None
    assert rule.events is not None
    assert rule.condition is not None
    assert rule.outcome is not None
    assert rule.options is not None
    assert rule.options.options["case_sensitive"] is False
    assert rule.options.options["max_events"] == 10
    assert rule.options.options["output_format"] == "json"


def test_enhanced_parser_parses_reference_and_regex_values() -> None:
    yaral_code = """
    rule enhanced_refs {
        events:
            $e.metadata.event_type matches /AUTH.*/im
            $e.metadata.product_name = %products%

        condition:
            principal.ip in %watchlist%
    }
    """

    parser = EnhancedYaraLParser(dedent(yaral_code))
    ast = parser.parse()

    assert parser.errors == []
    rule = ast.rules[0]
    events = rule.events
    assert events is not None
    assert any(
        isinstance(stmt.value, RegexPattern) for stmt in events.statements if hasattr(stmt, "value")
    )
    assert any(
        isinstance(stmt.value, ReferenceList)
        for stmt in events.statements
        if hasattr(stmt, "value")
    )


def test_enhanced_parser_error_recovery_keeps_parsing() -> None:
    yaral_code = """
    rule bad_rule {
        events:
            $e.metadata.event_type = "LOGIN"
        condition:
            #e >
    }

    rule good_rule {
        events:
            $e.metadata.event_type = "OK"
        condition:
            #e > 0
    }
    """

    parser = EnhancedYaraLParser(dedent(yaral_code))
    ast = parser.parse()

    assert parser.errors
    assert any(rule.name == "good_rule" for rule in ast.rules)



"""Folding ranges provider for YARAAST LSP."""

from __future__ import annotations

import logging
from typing import Any

from lsprotocol.types import FoldingRange, FoldingRangeKind

from yaraast.ast.rules import Rule
from yaraast.lsp.parsing import parse_for_lsp
from yaraast.lsp.structure import find_rule_end, find_rule_line, find_section_range

logger = logging.getLogger(__name__)


class FoldingRangesProvider:
    """Provide folding ranges for YARA code."""

    def get_folding_ranges(self, text: str) -> list[FoldingRange]:
        """Get all folding ranges in the document."""
        try:
            ast = parse_for_lsp(text)

            ranges = []

            # Add folding ranges for imports block
            if ast.imports:
                import_lines = self._get_import_block_lines(text, ast.imports)
                if import_lines:
                    ranges.append(
                        FoldingRange(
                            start_line=import_lines[0],
                            end_line=import_lines[1],
                            kind=FoldingRangeKind.Imports,
                        )
                    )

            # Add folding ranges for each rule
            for rule in ast.rules:
                rule_range = self._get_rule_folding_range(text, rule)
                if rule_range:
                    ranges.append(rule_range)

                # Add nested folding ranges for rule sections
                section_ranges = self._get_section_folding_ranges(text, rule)
                ranges.extend(section_ranges)

            return ranges

        except Exception:
            logger.debug("Operation failed in %s", __name__, exc_info=True)
            # Fallback to regex-based folding
            return self._fallback_folding_ranges(text)

    def _get_import_block_lines(self, text: str, imports: list[Any]) -> tuple[int, int] | None:
        """Get the line range for the imports block."""
        if not imports:
            return None

        lines = text.split("\n")

        # Find first import
        first_line = None
        last_line = None

        for line_num, line in enumerate(lines):
            if line.strip().startswith("import"):
                if first_line is None:
                    first_line = line_num
                last_line = line_num

        if first_line is not None and last_line is not None and last_line > first_line:
            return (first_line, last_line)

        return None

    def _get_rule_folding_range(self, text: str, rule: Rule) -> FoldingRange | None:
        """Get folding range for entire rule."""
        lines = text.split("\n")

        start_line = find_rule_line(lines, rule.name)
        if start_line is None or start_line < 0:
            return None

        end_line = find_rule_end(lines, start_line)
        if end_line is None or end_line <= start_line or "}" not in lines[end_line]:
            return None

        return FoldingRange(start_line=start_line, end_line=end_line, kind=FoldingRangeKind.Region)

    def _get_section_folding_ranges(self, text: str, rule: Rule) -> list[FoldingRange]:
        """Get folding ranges for rule sections (meta, strings, condition)."""
        ranges: list[FoldingRange] = []
        lines = text.split("\n")

        rule_start = find_rule_line(lines, rule.name)
        if rule_start is None or rule_start < 0:
            return ranges
        rule_end = find_rule_end(lines, rule_start)

        # Look for meta: section
        if rule.meta:
            meta_range = self._find_section_range(lines, rule_start, rule_end, "meta")
            if meta_range:
                ranges.append(
                    FoldingRange(
                        start_line=meta_range[0],
                        end_line=meta_range[1],
                        kind=FoldingRangeKind.Region,
                    )
                )

        # Look for strings: section
        if rule.strings:
            strings_range = self._find_section_range(lines, rule_start, rule_end, "strings")
            if strings_range:
                ranges.append(
                    FoldingRange(
                        start_line=strings_range[0],
                        end_line=strings_range[1],
                        kind=FoldingRangeKind.Region,
                    )
                )

        # Look for condition: section
        if rule.condition:
            condition_range = self._find_section_range(lines, rule_start, rule_end, "condition")
            if condition_range:
                ranges.append(
                    FoldingRange(
                        start_line=condition_range[0],
                        end_line=condition_range[1],
                        kind=FoldingRangeKind.Region,
                    )
                )

        return ranges

    def _find_section_range(
        self,
        lines: list[str],
        start_line: int,
        rule_end_or_section_name: int | str,
        section_name: str | None = None,
    ) -> tuple[int, int] | None:
        if section_name is None:
            raw_section_name = str(rule_end_or_section_name).removesuffix(":")
            rule_end = find_rule_end(lines, start_line)
        else:
            raw_section_name = section_name
            rule_end = int(rule_end_or_section_name)
        range_ = find_section_range(lines, raw_section_name, start_line, rule_end)
        if range_ is None or range_.end.line <= range_.start.line:
            return None
        return (range_.start.line, range_.end.line)

    def _fallback_folding_ranges(self, text: str) -> list[FoldingRange]:
        """Fallback regex-based folding when AST parsing fails."""
        ranges = []
        lines = text.split("\n")

        brace_stack = []

        for line_num, line in enumerate(lines):
            line.strip()

            # Track opening braces
            if "{" in line:
                brace_stack.append(line_num)

            # Track closing braces
            if "}" in line and brace_stack:
                start_line = brace_stack.pop()
                if line_num - start_line > 0:  # At least 1 line to fold
                    ranges.append(
                        FoldingRange(
                            start_line=start_line, end_line=line_num, kind=FoldingRangeKind.Region
                        )
                    )

        return ranges



from __future__ import annotations

import pytest

from yaraast.lexer.tokens import TokenType as T
from yaraast.yaral.ast_nodes import BinaryCondition, ReferenceList
from yaraast.yaral.enhanced_parser import EnhancedYaraLParser
from yaraast.yaral.lexer import YaraLToken
from yaraast.yaral.tokens import YaraLTokenType


def _tok(tt: T, value, yt: YaraLTokenType | None = None) -> YaraLToken:
    return YaraLToken(type=tt, value=value, line=1, column=1, length=1, yaral_type=yt)


def _set_tokens(p: EnhancedYaraLParser, toks: list[YaraLToken]) -> None:
    p.tokens = [*toks, _tok(T.EOF, None, YaraLTokenType.EOF)]
    p.current = 0


def test_enhanced_reference_check_success_and_missing_list_error() -> None:
    p = EnhancedYaraLParser("")
    _set_tokens(
        p,
        [
            _tok(T.IDENTIFIER, "metadata"),
            _tok(T.DOT, "."),
            _tok(T.IDENTIFIER, "event_type"),
            _tok(T.IN, "in"),
            _tok(T.IDENTIFIER, "%blocked%", YaraLTokenType.REFERENCE_LIST),
        ],
    )

    cond = p._parse_reference_check()
    assert isinstance(cond, BinaryCondition)
    assert cond.operator == "in"
    assert isinstance(cond.right, ReferenceList)
    assert cond.right.name == "blocked"

    p2 = EnhancedYaraLParser("")
    _set_tokens(
        p2,
        [
            _tok(T.IDENTIFIER, "metadata"),
            _tok(T.DOT, "."),
            _tok(T.IDENTIFIER, "event_type"),
            _tok(T.IN, "in"),
            _tok(T.STRING, "oops"),
        ],
    )
    with pytest.raises(ValueError, match="Expected reference list"):
        p2._parse_reference_check()


def test_enhanced_primary_condition_reference_list_branch_is_shadowed() -> None:
    p = EnhancedYaraLParser("")
    _set_tokens(p, [_tok(T.IDENTIFIER, "%blocked%", YaraLTokenType.REFERENCE_LIST)])

    with pytest.raises(ValueError, match="Expected comparison operator"):
        p._parse_primary_condition()


def test_enhanced_primary_condition_rejects_invalid_token() -> None:
    p = EnhancedYaraLParser("")
    _set_tokens(p, [_tok(T.STRING, "oops")])

    with pytest.raises(ValueError, match="Expected condition expression"):
        p._parse_primary_condition()



"""Shared visitor base for YARA-L that stubs core YARA visitors."""

from __future__ import annotations

from typing import Any, TypeVar

from yaraast.visitor import ASTVisitor

T = TypeVar("T")


class YaraLVisitor[T](ASTVisitor[T]):
    """YARA-L visitor base that stubs standard YARA AST methods."""

    def visit_yara_file(self, node: Any) -> T:
        raise NotImplementedError

    def visit_import(self, node: Any) -> T:
        raise NotImplementedError

    def visit_include(self, node: Any) -> T:
        raise NotImplementedError

    def visit_rule(self, node: Any) -> T:
        raise NotImplementedError

    def visit_tag(self, node: Any) -> T:
        raise NotImplementedError

    def visit_string_definition(self, node: Any) -> T:
        raise NotImplementedError

    def visit_plain_string(self, node: Any) -> T:
        raise NotImplementedError

    def visit_hex_string(self, node: Any) -> T:
        raise NotImplementedError

    def visit_regex_string(self, node: Any) -> T:
        raise NotImplementedError

    def visit_string_modifier(self, node: Any) -> T:
        raise NotImplementedError

    def visit_hex_token(self, node: Any) -> T:
        raise NotImplementedError

    def visit_hex_byte(self, node: Any) -> T:
        raise NotImplementedError

    def visit_hex_wildcard(self, node: Any) -> T:
        raise NotImplementedError

    def visit_hex_jump(self, node: Any) -> T:
        raise NotImplementedError

    def visit_hex_alternative(self, node: Any) -> T:
        raise NotImplementedError

    def visit_hex_nibble(self, node: Any) -> T:
        raise NotImplementedError

    def visit_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_identifier(self, node: Any) -> T:
        raise NotImplementedError

    def visit_string_identifier(self, node: Any) -> T:
        raise NotImplementedError

    def visit_string_wildcard(self, node: Any) -> T:
        raise NotImplementedError

    def visit_string_count(self, node: Any) -> T:
        raise NotImplementedError

    def visit_string_offset(self, node: Any) -> T:
        raise NotImplementedError

    def visit_string_length(self, node: Any) -> T:
        raise NotImplementedError

    def visit_integer_literal(self, node: Any) -> T:
        raise NotImplementedError

    def visit_double_literal(self, node: Any) -> T:
        raise NotImplementedError

    def visit_string_literal(self, node: Any) -> T:
        raise NotImplementedError

    def visit_regex_literal(self, node: Any) -> T:
        raise NotImplementedError

    def visit_boolean_literal(self, node: Any) -> T:
        raise NotImplementedError

    def visit_binary_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_unary_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_parentheses_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_set_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_range_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_function_call(self, node: Any) -> T:
        raise NotImplementedError

    def visit_array_access(self, node: Any) -> T:
        raise NotImplementedError

    def visit_member_access(self, node: Any) -> T:
        raise NotImplementedError

    def visit_condition(self, node: Any) -> T:
        raise NotImplementedError

    def visit_for_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_for_of_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_at_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_in_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_of_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_meta(self, node: Any) -> T:
        raise NotImplementedError

    def visit_module_reference(self, node: Any) -> T:
        raise NotImplementedError

    def visit_dictionary_access(self, node: Any) -> T:
        raise NotImplementedError

    def visit_comment(self, node: Any) -> T:
        raise NotImplementedError

    def visit_comment_group(self, node: Any) -> T:
        raise NotImplementedError

    def visit_defined_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_string_operator_expression(self, node: Any) -> T:
        raise NotImplementedError

    def visit_extern_rule(self, node: Any) -> T:
        raise NotImplementedError

    def visit_extern_rule_reference(self, node: Any) -> T:
        raise NotImplementedError

    def visit_extern_import(self, node: Any) -> T:
        raise NotImplementedError

    def visit_extern_namespace(self, node: Any) -> T:
        raise NotImplementedError

    def visit_pragma(self, node: Any) -> T:
        raise NotImplementedError

    def visit_in_rule_pragma(self, node: Any) -> T:
        raise NotImplementedError

    def visit_pragma_block(self, node: Any) -> T:
        raise NotImplementedError



"""Real coverage tests for generated protobuf module."""

from __future__ import annotations

import importlib
import os
import sys


def _purge_protobuf_and_serialization_modules() -> None:
    for name in list(sys.modules):
        if name.startswith("google.protobuf"):
            sys.modules.pop(name, None)
        if name.startswith("yaraast.serialization"):
            sys.modules.pop(name, None)


def test_yara_ast_pb2_python_impl_executes_generated_descriptor_block() -> None:
    # Force pure-python protobuf implementation so generated assignment block runs.
    os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"
    _purge_protobuf_and_serialization_modules()

    pb2 = importlib.import_module("yaraast.serialization.yara_ast_pb2")

    # Validate generated descriptor metadata is populated.
    assert pb2.DESCRIPTOR is not None
    assert pb2._YARAFILE._serialized_start > 0
    assert pb2._RULE._serialized_end > pb2._RULE._serialized_start

    # Exercise a broad set of generated messages/oneofs.
    yf = pb2.YaraFile()
    yf.imports.add(module="pe", alias="p")
    yf.includes.add(path="common.yar")
    rule = yf.rules.add(name="r")
    rule.modifiers.append("private")
    rule.tags.add(name="tag")
    rule.meta["author"].string_value = "me"
    s = rule.strings.add(identifier="$a")
    s.plain.value = "abc"
    s.plain.modifiers.add(name="nocase")
    assert s.WhichOneof("string_type") == "plain"

    hx = rule.strings.add(identifier="$h")
    t1 = hx.hex.tokens.add()
    t1.byte.value = "90"
    t2 = hx.hex.tokens.add()
    t2.wildcard.SetInParent()
    t3 = hx.hex.tokens.add()
    t3.jump.min_jump = 1
    t3.jump.max_jump = 2
    t4 = hx.hex.tokens.add()
    t4.nibble.high = True
    t4.nibble.value = 10
    assert t4.WhichOneof("token_type") == "nibble"

    rx = rule.strings.add(identifier="$r")
    rx.regex.regex = "ab+"
    rx.regex.modifiers.add(name="wide")
    assert rx.WhichOneof("string_type") == "regex"

    expr = pb2.Expression()
    expr.binary_expression.left.boolean_literal.value = True
    expr.binary_expression.operator = "and"
    expr.binary_expression.right.defined_expression.expression.identifier.name = "x"
    rule.condition.CopyFrom(expr)
    assert rule.condition.WhichOneof("expression_type") == "binary_expression"

    # Ensure serialization round-trip works for full message.
    data = yf.SerializeToString()
    parsed = pb2.YaraFile()
    parsed.ParseFromString(data)
    assert parsed.rules[0].name == "r"

    # Keep current protobuf module state for this process; re-switching implementation
    # in-process is unsupported by protobuf runtime internals.
    os.environ.pop("PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION", None)



"""CLI validation tests using libyara (no mocks)."""

from __future__ import annotations

import tempfile
from pathlib import Path

import pytest
from click.testing import CliRunner

from yaraast.cli.main import cli
from yaraast.libyara import YARA_AVAILABLE


@pytest.mark.skipif(not YARA_AVAILABLE, reason="yara-python not available")
def test_validate_cross_and_roundtrip(tmp_path: Path) -> None:
    rule_text = """
rule test_rule {
    strings:
        $a = "hello"
    condition:
        $a
}
"""
    rule_path = tmp_path / "rule.yar"
    rule_path.write_text(rule_text.strip())

    with tempfile.NamedTemporaryFile(mode="wb", delete=False) as f:
        f.write(b"hello world")
        data_path = Path(f.name)

    runner = CliRunner()
    try:
        cross_result = runner.invoke(
            cli,
            ["validate", "cross", str(rule_path), str(data_path)],
        )
        assert cross_result.exit_code == 0
        assert "Validation PASSED" in cross_result.output

        roundtrip_result = runner.invoke(
            cli,
            ["validate", "roundtrip", str(rule_path), "-d", str(data_path)],
        )
        assert roundtrip_result.exit_code == 0
        assert "Round-trip PASSED" in roundtrip_result.output
    finally:
        data_path.unlink(missing_ok=True)



"""Additional tests for format reporting helpers."""

from __future__ import annotations

from pathlib import Path

from rich.console import Console

from yaraast.cli import fmt_reporting as fr


def test_display_format_issues_truncates_and_reports_extra_count() -> None:
    console = Console(record=True, width=120)

    fr.display_format_issues(console, [f"issue {i}" for i in range(7)])

    output = console.export_text()
    assert "issue 0" in output
    assert "issue 4" in output
    assert "... and 2 more issues" in output


def test_display_format_diff_handles_no_changes() -> None:
    console = Console(record=True, width=120)

    fr.display_format_diff(
        console,
        Path("sample.yar"),
        "rule a { condition: true }\n",
        "rule a { condition: true }\n\n",
    )

    assert "No formatting changes needed" in console.export_text()


def test_display_format_check_without_issue_details_and_same_file_result() -> None:
    console = Console(record=True, width=120)
    path = Path("sample.yar")

    fr.display_format_check(console, path, True, [])
    fr.display_format_result(console, path, path, "compact")

    output = console.export_text()
    assert "sample.yar needs formatting" in output
    assert "Formatted sample.yar (compact style)" in output


def test_display_format_error_and_diff_fallback_line() -> None:
    console = Console(record=True, width=120)

    fr.display_format_error(console, "boom")
    fr._print_diff_lines(console, [" unchanged line"])

    output = console.export_text()
    assert "boom" in output
    assert "unchanged line" in output



"""Real tests for YARA-X syntax adapter (no mocks)."""

from __future__ import annotations

from textwrap import dedent

from yaraast.parser import Parser
from yaraast.yarax.compatibility_checker import CompatibilityIssue
from yaraast.yarax.feature_flags import YaraXFeatures
from yaraast.yarax.syntax_adapter import YaraXSyntaxAdapter


def _parse_yara(code: str):
    return Parser().parse(dedent(code))


def test_syntax_adapter_removes_duplicate_modifiers() -> None:
    code = """
    private private rule dup_mods {
        condition:
            true
    }
    """
    ast = _parse_yara(code)

    adapter = YaraXSyntaxAdapter(YaraXFeatures.yarax_strict(), target="yarax")
    adapted, count = adapter.adapt(ast)

    assert count >= 1
    assert sum(1 for m in adapted.rules[0].modifiers if str(m) == "private") == 1


def test_syntax_adapter_pads_base64() -> None:
    code = """
    rule short_base64 {
        strings:
            $a = "aa" base64
        condition:
            $a
    }
    """
    ast = _parse_yara(code)

    features = YaraXFeatures.yarax_strict()
    features.minimum_base64_length = 4
    adapter = YaraXSyntaxAdapter(features, target="yarax")
    adapted, count = adapter.adapt(ast)

    assert count >= 1
    string_def = adapted.rules[0].strings[0]
    assert len(string_def.value) >= 4


def test_syntax_adapter_escapes_regex_braces() -> None:
    code = r"""
    rule bad_regex {
        strings:
            $a = /a{b/
        condition:
            $a
    }
    """
    ast = _parse_yara(code)

    adapter = YaraXSyntaxAdapter(YaraXFeatures.yarax_strict(), target="yarax")
    adapted, count = adapter.adapt(ast)

    assert count >= 1
    string_def = adapted.rules[0].strings[0]
    assert "\\{" in string_def.regex


def test_generate_migration_guide() -> None:
    adapter = YaraXSyntaxAdapter(YaraXFeatures.yarax_strict(), target="yarax")
    issues = [
        CompatibilityIssue("error", None, "duplicate_modifier", "Duplicate", ""),
        CompatibilityIssue("error", None, "base64_too_short", "Base64", ""),
        CompatibilityIssue("error", None, "unescaped_brace", "Unescaped", ""),
    ]
    guide = adapter.generate_migration_guide(issues)

    assert "Migration Guide" in guide
    assert "Duplicate Modifiers" in guide
    assert "Base64 Pattern Length" in guide
    assert "Regex Brace Escaping" in guide



"""Shared helpers for semantic validators."""

from __future__ import annotations

from typing import TYPE_CHECKING

from yaraast.types.type_environment import TypeEnvironment

if TYPE_CHECKING:
    from yaraast.ast.base import YaraFile
    from yaraast.ast.rules import Rule


KNOWN_MODULES: frozenset[str] = frozenset(
    {
        "pe",
        "elf",
        "math",
        "hash",
        "cuckoo",
        "dotnet",
        "time",
        "console",
        "string",
        "dex",
        "macho",
        "magic",
        "vt",
    }
)

BUILTIN_FUNCTION_ARITY: dict[str, tuple[int, int]] = {
    "uint8": (1, 1),
    "uint16": (1, 1),
    "uint32": (1, 1),
    "int8": (1, 1),
    "int16": (1, 1),
    "int32": (1, 1),
    "uint8be": (1, 1),
    "uint16be": (1, 1),
    "uint32be": (1, 1),
    "int8be": (1, 1),
    "int16be": (1, 1),
    "int32be": (1, 1),
    "uint16le": (1, 1),
    "uint32le": (1, 1),
    "int16le": (1, 1),
    "int32le": (1, 1),
}


def populate_env_for_file(ast: YaraFile, env: TypeEnvironment) -> None:
    """Populate a type environment with modules and strings from a file."""
    for imp in ast.imports:
        alias = imp.alias if imp.alias else imp.module
        env.add_module(alias, imp.module)

    for rule in ast.rules:
        populate_env_for_rule(rule, env)


def populate_env_for_rule(rule: Rule, env: TypeEnvironment) -> None:
    """Populate a type environment with strings from a rule."""
    for string_def in rule.strings:
        env.add_string(string_def.identifier)



"""More tests for condition builder (no mocks)."""

from __future__ import annotations

import pytest

from yaraast.ast.conditions import ForExpression, InExpression
from yaraast.ast.expressions import BinaryExpression, ParenthesesExpression
from yaraast.builder.condition_builder import ConditionBuilder
from yaraast.errors import ValidationError


def test_condition_builder_basic_ops() -> None:
    expr = ConditionBuilder().string("$a").at(0).and_(ConditionBuilder().filesize().gt(10)).build()
    assert isinstance(expr, BinaryExpression)

    expr = ConditionBuilder().string("$a").in_range(0, 10).build()
    assert isinstance(expr, InExpression)


def test_condition_builder_group_and_for() -> None:
    grouped = ConditionBuilder().integer(1).eq(1).group().build()
    assert isinstance(grouped, ParenthesesExpression)

    iterable = ConditionBuilder().range(0, 3)
    body = ConditionBuilder().identifier("i").lt(2)
    loop = ConditionBuilder().for_any("i", iterable, body).build()
    assert isinstance(loop, ForExpression)


def test_condition_builder_errors_on_empty() -> None:
    with pytest.raises(ValidationError):
        ConditionBuilder().and_(ConditionBuilder().true())

    with pytest.raises(ValidationError):
        ConditionBuilder().group()

    with pytest.raises(ValidationError):
        ConditionBuilder().build()

    with pytest.raises(TypeError):
        ConditionBuilder().integer(1)._to_expression(3.14)  # type: ignore[arg-type]



"""Document lifecycle and runtime handler registration for the language server."""

from __future__ import annotations

import logging

from yaraast.lsp.lsp_types import (
    TEXT_DOCUMENT_DID_CHANGE,
    TEXT_DOCUMENT_DID_CLOSE,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_SAVE,
    WORKSPACE_DID_CHANGE_CONFIGURATION,
    WORKSPACE_DID_CHANGE_WATCHED_FILES,
    YARAAST_RUNTIME_STATUS,
    DidChangeConfigurationParams,
    DidChangeTextDocumentParams,
    DidChangeWatchedFilesParams,
    DidCloseTextDocumentParams,
    DidOpenTextDocumentParams,
    DidSaveTextDocumentParams,
)
from yaraast.lsp.server_feature_helpers import get_diagnostics, get_document_source

logger = logging.getLogger(__name__)


def register_document_handlers(server) -> None:
    # Length is due to 7 short handler registrations; splitting would reduce locality.
    @server.feature(TEXT_DOCUMENT_DID_OPEN)
    async def did_open(ls, params: DidOpenTextDocumentParams) -> None:
        ls.show_message_log("Document opened")
        runtime = getattr(ls, "runtime", None)
        if runtime is not None:
            runtime.open_document(
                params.text_document.uri,
                params.text_document.text,
                getattr(params.text_document, "version", None),
            )
        diagnostics = get_diagnostics(ls, params.text_document.text, params.text_document.uri)
        ls.publish_diagnostics(params.text_document.uri, diagnostics)

    @server.feature(TEXT_DOCUMENT_DID_CHANGE)
    async def did_change(ls, params: DidChangeTextDocumentParams) -> None:
        uri = params.text_document.uri
        changes = getattr(params, "content_changes", None)
        if changes:
            text = changes[-1].text
        else:
            try:
                text = ls.workspace.get_text_document(uri).source
            except Exception:
                logger.debug("Operation failed in %s", __name__, exc_info=True)
                text = ""
        runtime = getattr(ls, "runtime", None)
        if runtime is not None:
            runtime.update_document(uri, text, getattr(params.text_document, "version", None))
            if hasattr(runtime, "should_debounce") and runtime.should_debounce(
                uri, "push_diagnostics"
            ):
                return
        diagnostics = get_diagnostics(ls, text, uri)
        ls.publish_diagnostics(uri, diagnostics)

    @server.feature(TEXT_DOCUMENT_DID_SAVE)
    async def did_save(ls, params: DidSaveTextDocumentParams) -> None:
        uri = params.text_document.uri
        text = getattr(params, "text", None)
        if text is None:
            text = get_document_source(ls, uri)
        runtime = getattr(ls, "runtime", None)
        if runtime is not None:
            runtime.save_document(uri, text)
        diagnostics = get_diagnostics(ls, text, uri)
        ls.publish_diagnostics(uri, diagnostics)

    @server.feature(TEXT_DOCUMENT_DID_CLOSE)
    async def did_close(ls, params: DidCloseTextDocumentParams) -> None:
        ls.show_message_log("Document closed")
        runtime = getattr(ls, "runtime", None)
        if runtime is not None:
            runtime.close_document(params.text_document.uri)

    @server.feature(WORKSPACE_DID_CHANGE_CONFIGURATION)
    async def did_change_configuration(ls, params: DidChangeConfigurationParams) -> None:
        runtime = getattr(ls, "runtime", None)
        if runtime is not None:
            runtime.update_config(getattr(params, "settings", {}))

    @server.feature(WORKSPACE_DID_CHANGE_WATCHED_FILES)
    async def did_change_watched_files(ls, params: DidChangeWatchedFilesParams) -> None:
        runtime = getattr(ls, "runtime", None)
        if runtime is not None:
            runtime.handle_watched_files(getattr(params, "changes", []))

    @server.feature(YARAAST_RUNTIME_STATUS)
    async def runtime_status(ls, _params: object | None = None) -> dict[str, object]:
        runtime = getattr(ls, "runtime", None)
        if runtime is None:
            return {"available": False}
        return {"available": True, **runtime.get_status()}



"""Reporting helpers for roundtrip CLI output."""

from __future__ import annotations

from pathlib import Path
from typing import Any

import click


def _display_test_success(input_file: Path, result: dict[str, Any], format: str) -> None:
    """Display successful test results."""
    click.echo(f"✅ Round-trip test PASSED for {input_file}")
    click.echo(f"   Format: {format.upper()}")
    click.echo(f"   Original rules: {result['metadata']['original_rule_count']}")
    click.echo(
        f"   Reconstructed rules: {result['metadata']['reconstructed_rule_count']}",
    )


def _display_test_failure(input_file: Path, result: dict[str, Any], verbose: bool) -> None:
    """Display failed test results."""
    click.echo(f"❌ Round-trip test FAILED for {input_file}")
    click.echo(f"   Differences found: {len(result['differences'])}")

    if verbose:
        click.echo("\nDifferences:")
        for diff in result["differences"]:
            click.echo(f"   • {diff}")


def _display_verbose_source(result: dict[str, Any]) -> None:
    """Display verbose source comparison."""
    click.echo(
        f"\nOriginal source ({len(result['original_source'].splitlines())} lines):",
    )
    for i, line in enumerate(result["original_source"].splitlines()[:10], 1):
        click.echo(f"   {i:2d}: {line}")
    if len(result["original_source"].splitlines()) > 10:
        click.echo("      ... (truncated)")

    click.echo(
        f"\nReconstructed source ({len(result['reconstructed_source'].splitlines())} lines):",
    )
    for i, line in enumerate(result["reconstructed_source"].splitlines()[:10], 1):
        click.echo(f"   {i:2d}: {line}")
    if len(result["reconstructed_source"].splitlines()) > 10:
        click.echo("      ... (truncated)")


def display_serialize_result(
    output: Path | None,
    fmt: str,
    ast: Any,
    preserve_comments: bool,
    preserve_formatting: bool,
    serialized: str,
) -> None:
    """Display serialize results."""
    if output:
        click.echo(f"✅ Serialized to {output}")
        click.echo(f"   Format: {fmt.upper()}")
        click.echo(f"   Rules: {len(ast.rules)}")
        click.echo(f"   Comments preserved: {preserve_comments}")
        click.echo(f"   Formatting preserved: {preserve_formatting}")
    else:
        click.echo(serialized)


def display_deserialize_result(
    output: Path | None,
    fmt: str,
    ast: Any,
    preserve_formatting: bool,
    yara_code: str,
) -> None:
    """Display deserialize results."""
    if output:
        click.echo(f"✅ Generated YARA code to {output}")
        click.echo(f"   Rules: {len(ast.rules)}")
        click.echo(f"   Formatting preserved: {preserve_formatting}")
    else:
        click.echo(yara_code)


def display_pretty_result(
    output: Path | None,
    style: str,
    ast: Any,
    indent_size: int,
    max_line_length: int,
    formatted_code: str,
) -> None:
    """Display pretty print results."""
    if output:
        click.echo(f"✅ Pretty printed to {output}")
        click.echo(f"   Style: {style}")
        click.echo(f"   Rules: {len(ast.rules)}")
        click.echo(f"   Indent size: {indent_size}")
        click.echo(f"   Max line length: {max_line_length}")
    else:
        click.echo(formatted_code)


def display_pipeline_result(
    output: Path | None,
    yaml_content: str,
    include_manifest: bool,
    manifest_content: str | None,
    ast: Any,
) -> None:
    """Display pipeline serialization results."""
    if output:
        click.echo(f"✅ Pipeline YAML written to {output}")
    else:
        click.echo(yaml_content)

    if include_manifest:
        if output:
            click.echo(f"✅ Rules manifest written to {output.with_suffix('.manifest.yaml')}")
        else:
            click.echo("\n--- Rules Manifest ---")
            if manifest_content is not None:
                click.echo(manifest_content)

    click.echo("\nStatistics:")
    click.echo(f"   Rules: {len(ast.rules)}")
    click.echo(f"   Imports: {len(ast.imports)}")
    click.echo(f"   Includes: {len(ast.includes)}")



"""More tests for serialize reporting helpers (no mocks)."""

from __future__ import annotations

from pathlib import Path

from rich.console import Console

from yaraast.cli import serialize_reporting as sr


def test_display_export_import_and_diff_messages(tmp_path: Path) -> None:
    console = Console(record=True, width=120)

    sr.display_export_result(console, '{\n  "a":1\n}', "json", output=None, pretty=True, stats=None)
    sr.display_export_result(
        console,
        None,
        "protobuf",
        output=None,
        pretty=False,
        stats={
            "binary_size_bytes": 10,
            "text_size_bytes": 20,
            "compression_ratio": 2.0,
            "rules_count": 1,
            "imports_count": 0,
        },
    )
    sr.display_export_result(console, None, "json", output="out.json", pretty=False, stats=None)

    from yaraast.ast.base import YaraFile
    from yaraast.ast.expressions import BooleanLiteral
    from yaraast.ast.rules import Import, Rule

    ast = YaraFile(
        imports=[Import("pe")],
        rules=[
            Rule(name="a", condition=BooleanLiteral(True)),
            Rule(name="b", condition=BooleanLiteral(True)),
        ],
    )
    sr.display_import_result(console, "in.json", "json", ast, output=None)
    sr.display_import_result(console, "in.json", "json", ast, output="out.yar")

    sr.display_diff_no_changes(console)
    sr.display_diff_saved(console, "diff.patch", patch=True)
    sr.display_diff_saved(console, "diff.json", patch=False)

    out = console.export_text()
    assert "AST serialized successfully" in out
    assert "AST exported to out.json" in out
    assert "AST imported from in.json" in out
    assert "Code generation not yet implemented" in out
    assert "No differences found" in out
    assert "Patch file created" in out
    assert "Diff saved to" in out


def test_write_diff_output_and_display_info(tmp_path: Path) -> None:
    diff_data = {"changes": [{"path": "/r/1"}], "summary": {"added": 1}}
    json_out = tmp_path / "d.json"
    yaml_out = tmp_path / "d.yaml"

    sr.write_diff_output(str(json_out), "json", diff_data)
    sr.write_diff_output(str(yaml_out), "yaml", diff_data)

    assert '"changes"' in json_out.read_text(encoding="utf-8")
    assert "changes:" in yaml_out.read_text(encoding="utf-8")

    console = Console(record=True, width=120)
    info_data = {
        "rule_samples": ["a", "b", "c", "d"],
        "rule_count": 4,
        "import_count": 1,
        "include_count": 1,
        "import_list": ["pe"],
        "include_list": ["common.yar"],
        "rule_details": [
            {"name": "a", "strings": 1, "tags": 0, "meta": 1, "modifiers": "none"},
            {"name": "b", "strings": 2, "tags": 1, "meta": 0, "modifiers": "private"},
        ],
        "has_more_rules": True,
        "ast_hash": "hash123",
    }

    sr.display_info(console, str(tmp_path / "in.yar"), info_data)
    text = console.export_text()
    assert "AST Information" in text
    assert "Rule Analysis" in text
    assert "hash123" in text



"""Test full YARA language coverage features."""

import pytest

from yaraast import CodeGenerator, Parser
from yaraast.types import TypeValidator


def test_import_with_alias() -> None:
    """Test import statements with aliases."""
    yara_code = """
    import "pe" as windows
    import "elf" as linux
    import "math"

    rule test_imports {
        condition:
            windows.machine == 0x14c and
            linux.type == 0x02 and
            math.entropy(0, filesize) > 7.0
    }
    """

    parser = Parser()
    ast = parser.parse(yara_code)

    # Check imports
    assert len(ast.imports) == 3
    assert ast.imports[0].module == "pe"
    assert ast.imports[0].alias == "windows"
    assert ast.imports[1].module == "elf"
    assert ast.imports[1].alias == "linux"
    assert ast.imports[2].module == "math"
    assert ast.imports[2].alias is None

    # Generate code and verify
    generator = CodeGenerator()
    output = generator.generate(ast)
    assert 'import "pe" as windows' in output
    assert 'import "elf" as linux' in output
    assert 'import "math"' in output
    assert "as math" not in output


def test_regex_matches_regex() -> None:
    """Test regex matching another regex expression."""
    yara_code = """
    rule regex_matches_regex {
        condition:
            /foo.*bar/i matches /b[a-z]+r/
    }
    """

    parser = Parser()
    ast = parser.parse(yara_code)

    # Get the condition
    rule = ast.rules[0]
    condition = rule.condition

    # Verify it's a binary expression with matches operator
    assert condition.operator == "matches"

    # Check left side is regex literal
    assert condition.left is not None
    assert hasattr(condition.left, "pattern")
    assert condition.left.pattern == "foo.*bar"
    assert condition.left.modifiers == "i"

    # Check right side is regex literal
    assert condition.right is not None
    assert hasattr(condition.right, "pattern")
    assert condition.right.pattern == "b[a-z]+r"
    assert condition.right.modifiers == ""

    # Generate code
    generator = CodeGenerator()
    output = generator.generate(ast)
    assert "/foo.*bar/i matches /b[a-z]+r/" in output


def test_big_endian_little_endian_functions() -> None:
    """Test big-endian and little-endian integer functions."""
    yara_code = """
    rule endian_functions {
        condition:
            uint16be(0) == 0x4D5A and
            uint32be(4) == 0x50450000 and
            int16be(8) < 0 and
            int32be(12) > 1000 and
            uint16le(16) != 0x5A4D and
            uint32le(20) == 0x00004550 and
            int16le(24) >= -100 and
            int32le(28) <= 2000
    }
    """

    parser = Parser()
    ast = parser.parse(yara_code)

    # Type check should pass
    is_valid, errors = TypeValidator.validate(ast)
    assert is_valid, f"Type validation failed: {errors}"

    # Generate code
    generator = CodeGenerator()
    output = generator.generate(ast)

    # Check all functions are present
    for func in [
        "uint16be",
        "uint32be",
        "int16be",
        "int32be",
        "uint16le",
        "uint32le",
        "int16le",
        "int32le",
    ]:
        assert func in output


def test_complex_regex_in_expressions() -> None:
    """Test complex regex patterns in expressions."""
    yara_code = r"""
    rule complex_regex {
        strings:
            $a = /[a-z]{3,5}/ nocase
            $b = /\w+@\w+\.\w+/
            $c = /(?:https?|ftp):\/\/[^\s]+/i

        condition:
            $a and
            $b matches /.*@(gmail|yahoo|hotmail)\.com/ and
            /test\d+/ matches /test[0-9]+/
    }
    """

    parser = Parser()
    ast = parser.parse(yara_code)

    # Check the rule parsed correctly
    rule = ast.rules[0]
    assert len(rule.strings) == 3

    # Generate code
    generator = CodeGenerator()
    output = generator.generate(ast)

    # Verify regex patterns are preserved
    assert r"/[a-z]{3,5}/" in output
    assert r"/\w+@\w+\.\w+/" in output
    assert r"$b matches /.*@(gmail|yahoo|hotmail)\.com/" in output
    assert r"/test\d+/ matches /test[0-9]+/" in output


def test_string_matches_dynamic_regex() -> None:
    """Test string identifier matching dynamic regex."""
    yara_code = r"""
    rule string_matches_regex {
        strings:
            $email = /[a-z]+@[a-z]+\.com/
            $pattern = "malware"

        condition:
            $pattern matches /mal[a-z]+/ and
            "static_string" matches /static.*/ and
            $email matches /.*@evil\.com/
    }
    """

    parser = Parser()
    ast = parser.parse(yara_code)

    # Type validation should pass
    is_valid, errors = TypeValidator.validate(ast)
    assert is_valid, f"Type validation failed: {errors}"

    # Generate and verify
    generator = CodeGenerator()
    output = generator.generate(ast)
    assert "$pattern matches /mal[a-z]+/" in output
    assert '"static_string" matches /static.*/' in output
    assert r"$email matches /.*@evil\.com/" in output


def test_mixed_features() -> None:
    """Test combination of all new features."""
    yara_code = r"""
    import "pe" as peformat
    import "math" as m

    rule advanced_detection {
        meta:
            description = "Advanced detection using new features"

        strings:
            $mz = { 4D 5A }
            $str = "malicious"
            $url = /https?:\/\/[a-z0-9\.\-]+\/[a-z0-9]+/i

        condition:
            $mz at 0 and
            uint16be(0) == 0x4D5A and
            uint32le(uint32(0x3c)) == 0x00004550 and
            peformat.machine == 0x14c and
            m.entropy(0, 1024) > 6.5 and
            $str matches /mal[a-z]+/ and
            $url matches /.*\/download\//i and
            /[A-Z]{5,}/ matches /[A-Z]+/
    }
    """

    parser = Parser()
    ast = parser.parse(yara_code)

    # Should parse without errors
    assert len(ast.imports) == 2
    assert ast.imports[0].alias == "peformat"
    assert ast.imports[1].alias == "m"

    # Type check
    is_valid, errors = TypeValidator.validate(ast)
    assert is_valid, f"Type validation failed: {errors}"

    # Generate code
    generator = CodeGenerator()
    output = generator.generate(ast)

    # Verify all features are present
    assert 'import "pe" as peformat' in output
    assert 'import "math" as m' in output
    assert "uint16be(0) == 0x4D5A" in output
    assert "uint32le(uint32(0x3c))" in output
    assert "peformat.machine" in output
    assert "m.entropy" in output
    assert "$str matches /mal[a-z]+/" in output
    assert "/[A-Z]{5,}/ matches /[A-Z]+/" in output


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""Shared CLI helpers."""

from __future__ import annotations

import json
from pathlib import Path

from rich.console import Console
from rich.markup import escape

from yaraast.cli.parser_helpers import create_parser


def read_text(path: str | Path) -> str:
    """Read a text file with UTF-8 encoding."""
    return Path(path).read_text(encoding="utf-8")


def write_text(path: str | Path, content: str) -> None:
    """Write a text file with UTF-8 encoding."""
    Path(path).write_text(content, encoding="utf-8")


def write_json(path: str | Path, data: object, indent: int = 2) -> None:
    """Write JSON to disk with UTF-8 encoding."""
    write_text(path, json.dumps(data, indent=indent))


def format_json(
    data: object,
    indent: int | None = 2,
    *,
    sort_keys: bool | None = None,
    default: object | None = None,
    ensure_ascii: bool | None = None,
) -> str:
    """Format data as JSON with shared defaults."""
    kwargs: dict[str, object] = {}
    if sort_keys is not None:
        kwargs["sort_keys"] = sort_keys
    if default is not None:
        kwargs["default"] = default
    if ensure_ascii is not None:
        kwargs["ensure_ascii"] = ensure_ascii
    return json.dumps(data, indent=indent, **kwargs)


def parse_yara_text(source: str) -> object:
    """Parse YARA source into an AST."""
    return create_parser().parse(source)


def parse_yara_file(path: str | Path) -> object:
    """Read + parse a YARA file into an AST."""
    return parse_yara_text(read_text(path))


def print_cli_error(console: Console, exc: Exception, prefix: str = "Error") -> None:
    """Render a CLI error message safely for Rich."""
    console.print(f"[red]{prefix}: {escape(str(exc))}[/red]")



"""Real tests for LSP document highlight (no mocks)."""

from __future__ import annotations

from lsprotocol.types import Position

from yaraast.lsp.document_highlight import DocumentHighlightProvider


def _pos(line: int, char: int) -> Position:
    return Position(line=line, character=char)


def test_document_highlight_string_identifier() -> None:
    text = """
rule r {
    strings:
        $a = "abc"
    condition:
        $a and #a > 0 and @a == 10 and !a
}
""".lstrip()

    provider = DocumentHighlightProvider()
    highlights = provider.get_highlights(text, _pos(4, 8))
    assert len(highlights) >= 4


def test_document_highlight_identifier() -> None:
    text = """
rule alpha { condition: true }
rule beta { condition: alpha }
""".lstrip()

    provider = DocumentHighlightProvider()
    highlights = provider.get_highlights(text, _pos(1, 25))
    assert len(highlights) >= 2



"""Shared helpers for BaseVisitor."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Generic, Protocol, TypeVar, cast

from yaraast.ast.base import ASTNode
from yaraast.visitor.visitor import ASTVisitor

T = TypeVar("T", covariant=True)


class VisitorHelperProtocol(Protocol[T]):
    def _noop(self) -> T: ...

    def _visit_all(self, items: Sequence[ASTNode]) -> None: ...

    def _visit_if(self, node: ASTNode | None) -> None: ...


class BaseVisitorHelpersMixin(Generic[T]):  # noqa: UP046
    """Helper methods for BaseVisitor traversal."""

    def _default_visit(self, node: ASTNode) -> T:
        return cast(T, None)

    def _noop(self) -> T:
        return cast(T, None)

    def _visit_all(self, items: Sequence[ASTNode]) -> None:
        visitor = cast(ASTVisitor[T], self)
        for item in items:
            visitor.visit(item)

    def _visit_if(self, node: ASTNode | None) -> None:
        if node:
            visitor = cast(ASTVisitor[T], self)
            visitor.visit(node)



"""Job-oriented helpers for ParallelAnalyzer."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from yaraast.performance.parallel_job_helpers import (
    complete_job,
    export_graph_files,
    fail_job,
    parse_file_chunks,
    process_items,
    start_job,
)
from yaraast.performance.parallel_models import Job, JobStatus

if TYPE_CHECKING:
    from collections.abc import Callable

    from yaraast.ast.base import YaraFile
    from yaraast.performance.parallel_analyzer import ParallelAnalyzer


def analyze_complexity_parallel(
    analyzer: ParallelAnalyzer,
    asts: list[YaraFile],
    max_workers: int | None = None,
) -> list[Job]:
    """Create tracked complexity jobs for ASTs."""
    jobs = []
    for ast in asts:
        job = start_job("complexity")
        jobs.append(job)
        analyzer._stats["jobs_submitted"] += 1
        try:
            analysis = analyzer.analyze_file(ast, max_workers)
            analysis["metrics"] = analysis.get("stats", {})
            error_rate = analysis["stats"].get("errors", 0) / max(
                analysis["stats"].get("total_rules", 1),
                1,
            )
            analysis["quality_score"] = max(0, 100 - (error_rate * 100))
            complete_job(job, analysis)
            analyzer._stats["jobs_completed"] += 1
        except Exception as exc:
            fail_job(job, exc)
    return jobs


def generate_graphs_parallel(
    analyzer: ParallelAnalyzer,
    asts: list[YaraFile],
    output_dir: str | Path,
    graph_types: list[str] | None = None,
) -> list[Job]:
    """Generate dependency graph export jobs and update analyzer stats."""
    jobs = export_graph_files(asts, output_dir, graph_types)
    for job in jobs:
        analyzer._stats["jobs_submitted"] += 1
        if job.status == JobStatus.COMPLETED:
            analyzer._stats["jobs_completed"] += 1
    return jobs


def parse_files_parallel(
    analyzer: ParallelAnalyzer,
    file_paths: list,
    chunk_size: int = 10,
) -> list[Job]:
    """Parse file chunks into tracked jobs."""
    jobs = parse_file_chunks(file_paths, chunk_size)
    for job in jobs:
        analyzer._stats["jobs_submitted"] += 1
        if job.status == JobStatus.COMPLETED:
            analyzer._stats["jobs_completed"] += 1
    return jobs


def process_batch(
    analyzer: ParallelAnalyzer,
    items: list,
    worker_func: Callable,
    job_type: str = "batch",
    parameters: dict | None = None,
) -> list[Job]:
    """Process arbitrary items through tracked job helpers."""
    return process_items(items, worker_func, job_type=job_type, parameters=parameters)



"""More tests for CLI visitors (no mocks)."""

from __future__ import annotations

from yaraast.ast.expressions import RegexLiteral, StringLiteral
from yaraast.ast.modules import DictionaryAccess, ModuleReference
from yaraast.ast.operators import StringOperatorExpression
from yaraast.cli.visitors import ASTDumper


def test_ast_dumper_additional_nodes() -> None:
    dumper = ASTDumper()

    regex = RegexLiteral(pattern="ab.*", modifiers="i")
    regex_dump = dumper.visit_regex_literal(regex)
    assert regex_dump["pattern"] == "ab.*"

    mod_ref = ModuleReference(module="pe")
    mod_dump = dumper.visit_module_reference(mod_ref)
    assert mod_dump["module"] == "pe"

    dict_access = DictionaryAccess(object=mod_ref, key="CompanyName")
    dict_dump = dumper.visit_dictionary_access(dict_access)
    assert dict_dump["key"] == "CompanyName"

    string_op = StringOperatorExpression(
        left=StringLiteral(value="hello"),
        operator="icontains",
        right=StringLiteral(value="he"),
    )
    string_op_dump = dumper.visit_string_operator_expression(string_op)
    assert string_op_dump["operator"] == "icontains"



"""Tests for metrics and visualization functionality."""

import tempfile
from pathlib import Path

import pytest

from yaraast.metrics.complexity import ComplexityAnalyzer
from yaraast.metrics.complexity_model import ComplexityMetrics
from yaraast.metrics.html_tree import HtmlTreeGenerator
from yaraast.metrics.string_diagrams import StringDiagramGenerator

try:
    from yaraast.metrics.dependency_graph import DependencyGraphGenerator
except ModuleNotFoundError:
    DependencyGraphGenerator = None  # type: ignore[assignment]
from yaraast.parser import Parser


@pytest.fixture
def sample_yara_content() -> str:
    """Sample YARA content for testing."""
    return """
import "pe"
import "math"

rule TestRule : malware family {
    meta:
        author = "Test Author"
        description = "Test rule"
        version = "1.0"

    strings:
        $str1 = "Hello World" ascii wide
        $str2 = "Test String"
        $hex1 = { 48 65 6C 6C 6F ?? ?? 57 6F 72 6C 64 }
        $hex2 = { 90 90 [4-6] (AA | BB | CC) 90 90 }
        $regex1 = /test[0-9]+/i

    condition:
        pe.is_pe and (
            ($str1 or $str2) and
            for any of ($hex*) : ($ at pe.entry_point) and
            math.entropy(0, 1024) > 7.0
        )
}

private rule PrivateRule {
    strings:
        $a = "private test"

    condition:
        $a
}

global rule GlobalRule {
    condition:
        true
}
"""


@pytest.fixture
def parsed_ast(sample_yara_content):
    """Parse sample YARA content into AST."""
    parser = Parser()
    return parser.parse(sample_yara_content)


class TestComplexityAnalyzer:
    """Test complexity analysis functionality."""

    def test_basic_analysis(self, parsed_ast) -> None:
        """Test basic complexity analysis."""
        analyzer = ComplexityAnalyzer()
        metrics = analyzer.analyze(parsed_ast)

        assert isinstance(metrics, ComplexityMetrics)
        assert metrics.total_rules == 3
        assert metrics.total_imports == 2
        assert metrics.total_includes == 0
        assert metrics.private_rules == 1
        assert metrics.global_rules == 1
        assert metrics.rules_with_strings == 2
        assert metrics.rules_with_meta == 1
        assert metrics.rules_with_tags == 1

    def test_string_analysis(self, parsed_ast) -> None:
        """Test string complexity analysis."""
        analyzer = ComplexityAnalyzer()
        metrics = analyzer.analyze(parsed_ast)

        assert metrics.total_strings == 6  # TestRule (5) + PrivateRule (1)
        assert metrics.plain_strings == 3  # $str1, $str2, $a
        assert metrics.hex_strings == 2
        assert metrics.regex_strings == 1
        assert metrics.strings_with_modifiers > 0

        # Check hex pattern analysis
        assert metrics.hex_wildcards > 0
        # Note: hex_alternatives depends on parser support which may vary
        # assert metrics.hex_alternatives > 0

    def test_condition_complexity(self, parsed_ast) -> None:
        """Test condition complexity analysis."""
        analyzer = ComplexityAnalyzer()
        metrics = analyzer.analyze(parsed_ast)

        assert metrics.max_condition_depth > 0
        assert metrics.total_binary_ops > 0
        assert metrics.for_of_expressions > 0  # "for any of ($hex*)"
        assert len(metrics.cyclomatic_complexity) == 3  # One per rule

    def test_quality_scoring(self, parsed_ast) -> None:
        """Test quality scoring system."""
        analyzer = ComplexityAnalyzer()
        metrics = analyzer.analyze(parsed_ast)

        quality_score = metrics.get_quality_score()
        assert 0 <= quality_score <= 100

        grade = metrics.get_complexity_grade()
        assert grade in ["A", "B", "C", "D", "F"]

    def test_metrics_serialization(self, parsed_ast) -> None:
        """Test metrics to dictionary conversion."""
        analyzer = ComplexityAnalyzer()
        metrics = analyzer.analyze(parsed_ast)

        metrics_dict = metrics.to_dict()
        assert isinstance(metrics_dict, dict)
        assert "file_metrics" in metrics_dict
        assert "rule_metrics" in metrics_dict
        assert "string_metrics" in metrics_dict
        assert "condition_metrics" in metrics_dict
        assert "pattern_metrics" in metrics_dict
        assert "quality_metrics" in metrics_dict
        assert "dependencies" in metrics_dict

    def test_empty_file(self) -> None:
        """Test analysis of empty YARA file."""
        parser = Parser()
        ast = parser.parse("")

        analyzer = ComplexityAnalyzer()
        metrics = analyzer.analyze(ast)

        assert metrics.total_rules == 0
        assert metrics.total_strings == 0
        assert metrics.max_condition_depth == 0


@pytest.mark.skipif(
    DependencyGraphGenerator is None,
    reason="graphviz package is not installed",
)
class TestDependencyGraphGenerator:
    """Test dependency graph generation."""

    def test_basic_graph_generation(self, parsed_ast) -> None:
        """Test basic dependency graph generation."""
        generator = DependencyGraphGenerator()

        # Test that it returns GraphViz source
        graph_source = generator.generate_graph(parsed_ast)
        assert isinstance(graph_source, str)
        assert "digraph" in graph_source.lower()

    def test_rule_graph(self, parsed_ast) -> None:
        """Test rule-only dependency graph."""
        generator = DependencyGraphGenerator()

        graph_source = generator.generate_rule_graph(parsed_ast)
        assert isinstance(graph_source, str)
        assert "TestRule" in graph_source
        assert "PrivateRule" in graph_source
        assert "GlobalRule" in graph_source

    def test_module_graph(self, parsed_ast) -> None:
        """Test module dependency graph."""
        generator = DependencyGraphGenerator()

        graph_source = generator.generate_module_graph(parsed_ast)
        assert isinstance(graph_source, str)
        assert "pe" in graph_source
        assert "math" in graph_source

    def test_complexity_graph(self, parsed_ast) -> None:
        """Test complexity visualization graph."""
        generator = DependencyGraphGenerator()

        # Create mock complexity metrics
        complexity_metrics = {"TestRule": 8, "PrivateRule": 2, "GlobalRule": 1}

        graph_source = generator.generate_complexity_graph(
            parsed_ast,
            complexity_metrics,
        )
        assert isinstance(graph_source, str)
        assert "Complexity: 8" in graph_source
        assert "Complexity: 2" in graph_source
        assert "Complexity: 1" in graph_source

    def test_file_output(self, parsed_ast) -> None:
        """Test writing graph to file."""
        generator = DependencyGraphGenerator()

        with tempfile.TemporaryDirectory() as temp_dir:
            output_path = Path(temp_dir) / "test_graph.dot"
            result_path = generator.generate_graph(
                parsed_ast,
                str(output_path),
                "dot",
            )
            assert result_path.endswith(".dot")
            assert Path(result_path).exists()
            assert "digraph" in Path(result_path).read_text(encoding="utf-8")

    def test_dependency_stats(self, parsed_ast) -> None:
        """Test dependency statistics collection."""
        generator = DependencyGraphGenerator()
        generator.generate_graph(parsed_ast)  # Populate internal state

        stats = generator.get_dependency_stats()
        assert isinstance(stats, dict)
        assert "total_rules" in stats
        assert "total_imports" in stats
        assert "rules_with_strings" in stats
        assert stats["total_rules"] == 3
        assert stats["total_imports"] == 2


class TestHtmlTreeGenerator:
    """Test HTML tree visualization."""

    def test_basic_html_generation(self, parsed_ast) -> None:
        """Test basic HTML tree generation."""
        generator = HtmlTreeGenerator()

        html_content = generator.generate_html(parsed_ast)
        assert isinstance(html_content, str)
        assert "<!DOCTYPE html>" in html_content
        assert "YARA AST Visualization" in html_content
        assert "TestRule" in html_content

    def test_interactive_html(self, parsed_ast) -> None:
        """Test interactive HTML generation with search."""
        generator = HtmlTreeGenerator()

        html_content = generator.generate_interactive_html(
            parsed_ast,
            title="Test Interactive",
        )
        assert isinstance(html_content, str)
        assert "Test Interactive" in html_content
        assert "searchNodes" in html_content  # JavaScript search function
        assert "filterNodes" in html_content  # JavaScript filter function

    def test_html_file_output(self, parsed_ast) -> None:
        """Test writing HTML to file."""
        generator = HtmlTreeGenerator()

        with tempfile.TemporaryDirectory() as temp_dir:
            output_path = Path(temp_dir) / "test_tree.html"

            generator.generate_html(parsed_ast, str(output_path))

            assert output_path.exists()
            content = output_path.read_text()
            assert "TestRule" in content
            assert "<!DOCTYPE html>" in content

    def test_metadata_inclusion(self, parsed_ast) -> None:
        """Test metadata inclusion control."""
        # With metadata
        generator_with_meta = HtmlTreeGenerator(include_metadata=True)
        html_with_meta = generator_with_meta.generate_html(parsed_ast)

        # Without metadata
        generator_no_meta = HtmlTreeGenerator(include_metadata=False)
        html_no_meta = generator_no_meta.generate_html(parsed_ast)

        # Both should be valid HTML but different
        assert "<!DOCTYPE html>" in html_with_meta
        assert "<!DOCTYPE html>" in html_no_meta
        assert html_with_meta != html_no_meta

    def test_custom_title(self, parsed_ast) -> None:
        """Test custom page title."""
        generator = HtmlTreeGenerator()
        custom_title = "My Custom YARA Analysis"

        html_content = generator.generate_html(parsed_ast, title=custom_title)
        assert custom_title in html_content
        assert f"<title>{custom_title}</title>" in html_content


class TestStringDiagramGenerator:
    """Test string pattern diagram generation."""

    def test_pattern_flow_diagram(self, parsed_ast) -> None:
        """Test pattern flow diagram generation."""
        generator = StringDiagramGenerator()

        diagram_source = generator.generate_pattern_flow_diagram(parsed_ast)
        assert isinstance(diagram_source, str)
        assert "digraph" in diagram_source.lower()
        assert "Plain String Patterns" in diagram_source
        assert "Hex Patterns" in diagram_source
        assert "Regex Patterns" in diagram_source

    def test_pattern_complexity_diagram(self, parsed_ast) -> None:
        """Test pattern complexity visualization."""
        generator = StringDiagramGenerator()

        diagram_source = generator.generate_pattern_complexity_diagram(parsed_ast)
        assert isinstance(diagram_source, str)
        assert "Complexity" in diagram_source
        assert "neato" in diagram_source  # Should use neato engine

    def test_pattern_similarity_diagram(self, parsed_ast) -> None:
        """Test pattern similarity clustering."""
        generator = StringDiagramGenerator()

        diagram_source = generator.generate_pattern_similarity_diagram(parsed_ast)
        assert isinstance(diagram_source, str)
        assert "fdp" in diagram_source  # Should use fdp engine

    def test_hex_pattern_diagram(self, parsed_ast) -> None:
        """Test hex pattern analysis diagram."""
        generator = StringDiagramGenerator()

        diagram_source = generator.generate_hex_pattern_diagram(parsed_ast)
        assert isinstance(diagram_source, str)
        # Should contain analysis of hex patterns
        assert "$hex1" in diagram_source or "$hex2" in diagram_source

    def test_pattern_statistics(self, parsed_ast) -> None:
        """Test pattern statistics collection."""
        generator = StringDiagramGenerator()

        # Generate any diagram to populate internal state
        generator.generate_pattern_flow_diagram(parsed_ast)

        stats = generator.get_pattern_statistics()
        assert isinstance(stats, dict)
        assert "total_patterns" in stats
        assert "by_type" in stats
        assert "complexity_distribution" in stats
        assert stats["total_patterns"] == 6  # TestRule (5) + PrivateRule (1)
        assert stats["by_type"]["plain"] == 3  # $str1, $str2, $a
        assert stats["by_type"]["hex"] == 2
        assert stats["by_type"]["regex"] == 1

    def test_file_output(self, parsed_ast) -> None:
        """Test writing diagram to file."""
        generator = StringDiagramGenerator()

        with tempfile.TemporaryDirectory() as temp_dir:
            output_path = Path(temp_dir) / "test_patterns.dot"
            result_path = generator.generate_pattern_flow_diagram(
                parsed_ast,
                str(output_path),
                "dot",
            )
            assert result_path.endswith(".dot")
            assert Path(result_path).exists()
            assert "digraph" in Path(result_path).read_text(encoding="utf-8")

    def test_empty_patterns(self) -> None:
        """Test handling of files with no string patterns."""
        parser = Parser()
        ast = parser.parse(
            """
        rule EmptyRule {
            condition: true
        }
        """,
        )

        generator = StringDiagramGenerator()

        # Should handle gracefully
        diagram_source = generator.generate_pattern_flow_diagram(ast)
        assert isinstance(diagram_source, str)

        stats = generator.get_pattern_statistics()
        assert stats["total_patterns"] == 0


class TestMetricsIntegration:
    """Test integration between different metrics components."""

    def test_complexity_with_dependency_graph(self, parsed_ast) -> None:
        """Test using complexity metrics with dependency graph."""
        # Get complexity metrics
        analyzer = ComplexityAnalyzer()
        metrics = analyzer.analyze(parsed_ast)

        # Use with dependency graph
        generator = DependencyGraphGenerator()
        graph_source = generator.generate_complexity_graph(
            parsed_ast,
            metrics.cyclomatic_complexity,
        )

        assert isinstance(graph_source, str)
        assert "Complexity" in graph_source

    def test_all_metrics_on_same_ast(self, parsed_ast) -> None:
        """Test that all metrics work on the same AST."""
        # Complexity analysis
        complexity_analyzer = ComplexityAnalyzer()
        complexity_metrics = complexity_analyzer.analyze(parsed_ast)

        # Dependency graph
        dep_generator = DependencyGraphGenerator()
        dep_graph = dep_generator.generate_graph(parsed_ast)

        # HTML tree
        html_generator = HtmlTreeGenerator()
        html_tree = html_generator.generate_html(parsed_ast)

        # String patterns
        pattern_generator = StringDiagramGenerator()
        pattern_diagram = pattern_generator.generate_pattern_flow_diagram(parsed_ast)

        # All should succeed
        assert complexity_metrics.total_rules == 3
        assert isinstance(dep_graph, str)
        assert isinstance(html_tree, str)
        assert isinstance(pattern_diagram, str)

    def test_metrics_consistency(self, parsed_ast) -> None:
        """Test that different metrics report consistent counts."""
        # Complexity analysis
        complexity_analyzer = ComplexityAnalyzer()
        complexity_metrics = complexity_analyzer.analyze(parsed_ast)

        # Dependency analysis
        dep_generator = DependencyGraphGenerator()
        dep_generator.generate_graph(parsed_ast)  # Populate state
        dep_stats = dep_generator.get_dependency_stats()

        # Should report same rule counts
        assert complexity_metrics.total_rules == dep_stats["total_rules"]
        assert complexity_metrics.total_imports == dep_stats["total_imports"]

    def test_large_file_handling(self) -> None:
        """Test metrics on a larger, more complex YARA file."""
        # Create a more complex YARA file
        complex_yara = """
import "pe"
import "math"
import "hash"

rule ComplexRule1 : malware trojan {
    meta:
        author = "Test"
        family = "TestFamily"
    strings:
        $s1 = "test1" ascii wide nocase
        $s2 = "test2"
        $h1 = { 90 90 [1-3] AA BB ?? CC DD }
        $h2 = { (48 | 49 | 4A) 65 6C 6C 6F [4-8] 57 6F 72 6C 64 }
        $r1 = /test[a-z0-9]{1,10}/i
        $r2 = /^(GET|POST)\\s+/
    condition:
        pe.is_pe and pe.is_32bit() and (
            (2 of ($s*) and 1 of ($h*)) or
            (for any of ($r*) : ($ in (0..1024))) and
            (math.entropy(pe.entry_point, 256) > 6.5 or
             hash.md5(0, 1024) == "d41d8cd98f00b204e9800998ecf8427e")
        )
}

rule ComplexRule2 {
    strings:
        $a = { FF 25 [4] }
        $b = { E8 [4] }
    condition:
        for all i in (1..#a) : (
            @a[i] < pe.entry_point and
            for any j in (1..#b) : (
                @b[j] > @a[i] and @b[j] < @a[i] + 100
            )
        )
}

private rule HelperRule {
    condition:
        pe.number_of_sections > 3
}
        """

        parser = Parser()
        ast = parser.parse(complex_yara)

        # Run all metrics
        complexity_analyzer = ComplexityAnalyzer()
        metrics = complexity_analyzer.analyze(ast)

        dep_generator = DependencyGraphGenerator()
        dep_graph = dep_generator.generate_graph(ast)

        pattern_generator = StringDiagramGenerator()
        pattern_generator.get_pattern_statistics()

        # Verify complex analysis worked
        assert metrics.total_rules == 3
        assert metrics.max_condition_depth > 2  # Should be fairly deep
        assert len(metrics.cyclomatic_complexity) == 3
        assert isinstance(dep_graph, str)

        # Verify cyclomatic complexity was calculated per-rule
        assert all(v >= 1 for v in metrics.cyclomatic_complexity.values())



"""Enhanced pretty printer for YARA rules with formatting preservation."""

from __future__ import annotations

from dataclasses import dataclass
from io import StringIO
from typing import TYPE_CHECKING, Any

from yaraast.ast.strings import HexString, PlainString, RegexString, StringDefinition
from yaraast.codegen.comment_aware_generator import CommentAwareCodeGenerator
from yaraast.codegen.pretty_printer_helpers import (
    calculate_meta_alignment_column,
    calculate_string_alignment_column,
    expression_to_string,
)
from yaraast.codegen.pretty_printer_layout import (
    visit_rule as layout_visit_rule,
    visit_yara_file as layout_visit_yara_file,
    write_condition_section as layout_write_condition_section,
    write_string_definition as layout_write_string_definition,
)
from yaraast.codegen.pretty_printer_sections import (
    write_hex_string_aligned as section_write_hex_string_aligned,
    write_meta_entry as section_write_meta_entry,
    write_meta_section as section_write_meta_section,
    write_plain_string_aligned as section_write_plain_string_aligned,
    write_regex_string_aligned as section_write_regex_string_aligned,
    write_strings_section as section_write_strings_section,
    write_wrapped_condition as section_write_wrapped_condition,
)

if TYPE_CHECKING:
    from yaraast.ast.base import YaraFile
    from yaraast.ast.expressions import Expression
    from yaraast.ast.rules import Rule


@dataclass
class PrettyPrintOptions:
    """Options for pretty printing YARA rules."""

    # Indentation
    indent_size: int = 4
    indent_with_tabs: bool = False

    # Spacing
    blank_lines_before_rule: int = 2
    blank_lines_after_imports: int = 2
    blank_lines_after_includes: int = 1
    blank_lines_between_sections: int = 1
    space_around_operators: bool = True
    space_after_comma: bool = True

    # Alignment
    align_string_definitions: bool = True
    align_meta_values: bool = True
    align_comments: bool = True
    min_alignment_column: int = 40

    # Comments
    preserve_comments: bool = True
    comment_column: int = 60
    inline_comment_spacing: int = 2

    # String formatting
    quote_style: str = "double"  # "double", "single", "preserve"
    hex_uppercase: bool = True
    hex_spacing: bool = True

    # Line length and wrapping
    max_line_length: int = 120
    wrap_long_conditions: bool = True
    wrap_long_strings: bool = False

    # Sorting
    sort_imports: bool = True
    sort_includes: bool = True
    sort_meta_keys: bool = False
    sort_tags: bool = True

    # Style preferences
    compact_conditions: bool = False
    verbose_conditions: bool = False
    preserve_original_style: bool = False


class PrettyPrinter(CommentAwareCodeGenerator):
    """Enhanced pretty printer with advanced formatting options."""

    def __init__(self, options: PrettyPrintOptions | None = None) -> None:
        self.options = options or PrettyPrintOptions()
        super().__init__(
            indent_size=self.options.indent_size,
            preserve_comments=self.options.preserve_comments,
        )
        self._string_alignment_column = 0
        self._meta_alignment_column = 0

    def pretty_print(self, ast: YaraFile) -> str:
        """Pretty print the entire YARA file."""
        self.buffer = StringIO()
        self.indent_level = 0

        # Calculate alignment columns if needed
        if self.options.align_string_definitions:
            self._string_alignment_column = calculate_string_alignment_column(ast)
        if self.options.align_meta_values:
            self._meta_alignment_column = calculate_meta_alignment_column(
                ast,
                self.options.min_alignment_column,
            )

        return self.visit_yara_file(ast)

    def visit_yara_file(self, node: YaraFile) -> str:
        return layout_visit_yara_file(self, node)

    def visit_rule(self, node: Rule) -> str:
        return layout_visit_rule(self, node)

    def _write_meta_section(self, meta: dict[str, Any] | list[Any]) -> None:
        section_write_meta_section(self, meta)

    def _write_meta_entry(self, key: str, value: Any) -> None:
        section_write_meta_entry(self, key, value)

    def _write_strings_section(self, strings: list[StringDefinition]) -> None:
        section_write_strings_section(self, strings)

    def _write_string_definition(self, string_def: StringDefinition) -> None:
        layout_write_string_definition(self, string_def)

    def _write_plain_string_aligned(self, node: PlainString) -> None:
        section_write_plain_string_aligned(self, node)

    def _write_hex_string_aligned(self, node: HexString) -> None:
        section_write_hex_string_aligned(self, node)

    def _write_regex_string_aligned(self, node: RegexString) -> None:
        section_write_regex_string_aligned(self, node)

    def _write_condition_section(self, condition: Expression) -> None:
        layout_write_condition_section(self, condition)

    def _write_wrapped_condition(self, condition_str: str) -> None:
        section_write_wrapped_condition(self, condition_str)

    def _expression_to_string(self, expr: Expression) -> str:
        """Convert expression to string (simplified)."""
        # This is a simplified implementation
        # In practice, would use a separate visitor for expression serialization
        return expression_to_string(expr)


class StylePresets:
    """Predefined style presets for different use cases."""

    @staticmethod
    def compact() -> PrettyPrintOptions:
        """Compact style for minimal whitespace."""
        return PrettyPrintOptions(
            blank_lines_before_rule=1,
            blank_lines_after_imports=1,
            blank_lines_after_includes=0,
            blank_lines_between_sections=0,
            align_string_definitions=False,
            align_meta_values=False,
            compact_conditions=True,
            max_line_length=80,
        )

    @staticmethod
    def readable() -> PrettyPrintOptions:
        """Readable style with good spacing and alignment."""
        return PrettyPrintOptions(
            blank_lines_before_rule=2,
            blank_lines_after_imports=2,
            blank_lines_after_includes=1,
            blank_lines_between_sections=1,
            align_string_definitions=True,
            align_meta_values=True,
            align_comments=True,
            max_line_length=120,
        )

    @staticmethod
    def dense() -> PrettyPrintOptions:
        """Dense style for large files."""
        return PrettyPrintOptions(
            blank_lines_before_rule=1,
            blank_lines_after_imports=1,
            blank_lines_after_includes=0,
            blank_lines_between_sections=0,
            align_string_definitions=True,
            align_meta_values=False,
            max_line_length=100,
            compact_conditions=True,
        )

    @staticmethod
    def verbose() -> PrettyPrintOptions:
        """Verbose style with extensive spacing."""
        return PrettyPrintOptions(
            blank_lines_before_rule=3,
            blank_lines_after_imports=3,
            blank_lines_after_includes=2,
            blank_lines_between_sections=2,
            align_string_definitions=True,
            align_meta_values=True,
            align_comments=True,
            verbose_conditions=True,
            max_line_length=140,
        )


# Convenience functions
def pretty_print(ast: YaraFile, options: PrettyPrintOptions | None = None) -> str:
    """Pretty print YARA AST with specified options."""
    printer = PrettyPrinter(options)
    return printer.pretty_print(ast)


def pretty_print_compact(ast: YaraFile) -> str:
    """Pretty print with compact style."""
    return pretty_print(ast, StylePresets.compact())


def pretty_print_readable(ast: YaraFile) -> str:
    """Pretty print with readable style."""
    return pretty_print(ast, StylePresets.readable())


def pretty_print_dense(ast: YaraFile) -> str:
    """Pretty print with dense style."""
    return pretty_print(ast, StylePresets.dense())


def pretty_print_verbose(ast: YaraFile) -> str:
    """Pretty print with verbose style."""
    return pretty_print(ast, StylePresets.verbose())



"""Operational facade for the metrics subsystem."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any

from yaraast.ast.base import YaraFile
from yaraast.metrics.capabilities import CAPABILITIES, MetricsCapability, get_capability
from yaraast.metrics.complexity import ComplexityAnalyzer
from yaraast.metrics.dependency_graph import DependencyGraphGenerator
from yaraast.metrics.html_tree import HtmlTreeGenerator
from yaraast.metrics.string_diagrams import StringDiagramGenerator


@dataclass(frozen=True)
class MetricsSubsystem:
    """Provide a single conceptual entry point for metrics capabilities."""

    complexity_analyzer: type[ComplexityAnalyzer] = ComplexityAnalyzer
    dependency_graph_generator: type[DependencyGraphGenerator] = DependencyGraphGenerator
    html_tree_generator: type[HtmlTreeGenerator] = HtmlTreeGenerator
    string_diagram_generator: type[StringDiagramGenerator] = StringDiagramGenerator

    def new_complexity_analyzer(self) -> ComplexityAnalyzer:
        return self.complexity_analyzer()

    def new_dependency_graph_generator(self) -> DependencyGraphGenerator:
        return self.dependency_graph_generator()

    def new_html_tree_generator(self) -> HtmlTreeGenerator:
        return self.html_tree_generator()

    def new_string_diagram_generator(self) -> StringDiagramGenerator:
        return self.string_diagram_generator()

    def list_capabilities(self) -> tuple[MetricsCapability, ...]:
        return CAPABILITIES

    def get_capability(self, name: str) -> MetricsCapability | None:
        return get_capability(name)

    def analyze_complexity(self, ast: YaraFile) -> Any:
        from yaraast.metrics.workflows import analyze_complexity

        return analyze_complexity(ast)

    def build_report(
        self, ast: YaraFile, output_dir: Path, base_name: str, image_format: str
    ) -> Any:
        from yaraast.metrics.workflows import build_report

        return build_report(ast, output_dir, base_name, image_format)


METRICS = MetricsSubsystem()



"""Test AST-based analysis features."""

from yaraast.analysis import (
    AnalysisReport,
    BestPracticesAnalyzer,
    OptimizationAnalyzer,
    OptimizationReport,
)
from yaraast.parser import Parser


class TestBestPracticesAnalyzer:
    """Test best practices analysis."""

    def test_rule_naming_conventions(self) -> None:
        """Test rule name convention checking."""
        rule_text = """
        rule bad123name { condition: true }
        rule _also_bad { condition: true }
        rule good_name { condition: true }
        rule a { condition: true }  // too short
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = BestPracticesAnalyzer()
        report = analyzer.analyze(ast)

        # Check for naming issues
        warnings = report.get_by_severity("warning")
        assert len(warnings) >= 2  # bad names

        info = report.get_by_severity("info")
        assert any("descriptive" in s.message for s in info)  # short name

    def test_string_naming_conventions(self) -> None:
        """Test string identifier conventions."""
        rule_text = """
        rule test_strings {
            strings:
                $good = "test"
                $123 = "starts with number"
                $a = "ok"
            condition:
                any of them
        }
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = BestPracticesAnalyzer()
        report = analyzer.analyze(ast)

        warnings = report.get_by_severity("warning")
        assert any("$name convention" in s.message for s in warnings)

    def test_section_order_suggestion(self) -> None:
        """AST-only analyzer should not guess section order without source-order evidence."""
        rule_text = """
        rule wrong_order {
            condition:
                true
            strings:
                $a = "test"
            meta:
                author = "test"
        }
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = BestPracticesAnalyzer()
        report = analyzer.analyze(ast)

        info = report.get_by_severity("info")
        assert not any("section order" in s.message for s in info)

    def test_unused_strings_detection(self) -> None:
        """Test detection of unused strings."""
        rule_text = """
        rule unused_test {
            strings:
                $used = "test"
                $unused = "never referenced"
            condition:
                $used
        }
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = BestPracticesAnalyzer()
        report = analyzer.analyze(ast)

        warnings = report.get_by_severity("warning")
        assert any("$unused" in s.message and "never used" in s.message for s in warnings)

    def test_short_string_warning(self) -> None:
        """Test warning for very short strings."""
        rule_text = """
        rule short_strings {
            strings:
                $a = "ab"  // too short
                $b = "this is long enough"
            condition:
                all of them
        }
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = BestPracticesAnalyzer()
        report = analyzer.analyze(ast)

        info = report.get_by_severity("info")
        assert any("Short string" in s.message for s in info)

    def test_duplicate_detection(self) -> None:
        """Test detection of duplicate names."""
        rule_text = """
        rule dup_test {
            strings:
                $a = "first"
                $a = "duplicate"
            condition:
                $a
        }
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = BestPracticesAnalyzer()
        report = analyzer.analyze(ast)

        errors = report.get_by_severity("error")
        assert any("Duplicate string identifier" in s.message for s in errors)


class TestOptimizationAnalyzer:
    """Test optimization analysis."""

    def test_hex_pattern_consolidation(self) -> None:
        """Test suggestion for consolidating hex patterns."""
        rule_text = """
        rule hex_patterns {
            strings:
                $a = { 48 65 6c 6c 6f 20 }
                $b = { 48 65 6c 6c 6f 21 }
                $c = { 48 65 6c 6c 6f 3f }
            condition:
                any of them
        }
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = OptimizationAnalyzer()
        report = analyzer.analyze(ast)

        # Should suggest consolidation
        assert any("common prefix" in s.description for s in report.suggestions)

    def test_overlapping_patterns(self) -> None:
        """Test detection of overlapping patterns."""
        rule_text = """
        rule overlapping {
            strings:
                $a = "malicious"
                $b = "malicious code"  // contains $a
            condition:
                all of them
        }
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = OptimizationAnalyzer()
        report = analyzer.analyze(ast)

        # Should detect redundancy
        assert any("contained in" in s.description for s in report.suggestions)

    def test_redundant_comparisons(self) -> None:
        """Test detection of redundant comparisons."""
        rule_text = """
        rule redundant {
            strings:
                $a = "test"
            condition:
                #a > 5 and #a > 10  // second makes first redundant
        }
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = OptimizationAnalyzer()
        report = analyzer.analyze(ast)

        # Should detect redundancy
        assert any("Redundant comparison" in s.description for s in report.suggestions)

    def test_hex_wildcard_optimization(self) -> None:
        """Test warning for excessive wildcards."""
        rule_text = """
        rule wildcards {
            strings:
                $mostly_wildcards = { 48 ?? ?? ?? ?? ?? ?? ?? }
            condition:
                $mostly_wildcards
        }
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = BestPracticesAnalyzer()  # This check is in best practices
        report = analyzer.analyze(ast)

        warnings = report.get_by_severity("warning")
        assert any("wildcards" in s.message and "inefficient" in s.message for s in warnings)

    def test_cross_rule_duplication(self) -> None:
        """Test detection of patterns duplicated across rules."""
        rule_text = """
        rule rule1 {
            strings:
                $a = "same pattern"
            condition: $a
        }

        rule rule2 {
            strings:
                $b = "same pattern"
            condition: $b
        }

        rule rule3 {
            strings:
                $c = "same pattern"
            condition: $c
        }
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = OptimizationAnalyzer()
        report = analyzer.analyze(ast)

        # Should detect duplication
        assert any("Same plain pattern used in" in s.description for s in report.suggestions)

    def test_complex_condition_warning(self) -> None:
        """Test warning for overly complex conditions."""
        # Create deeply nested condition
        rule_text = """
        rule complex {
            condition:
                (true and (false or (true and (false or (true and false)))))
        }
        """

        parser = Parser()
        ast = parser.parse(rule_text)

        analyzer = OptimizationAnalyzer()
        report = analyzer.analyze(ast)

        # Should warn about complexity
        assert any("deep condition nesting" in s.description for s in report.suggestions)


class TestReportFeatures:
    """Test report functionality."""

    def test_analysis_report_categorization(self) -> None:
        """Test report categorization features."""
        report = AnalysisReport()

        report.add_suggestion("rule1", "style", "info", "Style issue")
        report.add_suggestion("rule2", "optimization", "warning", "Optimize this")
        report.add_suggestion("rule3", "structure", "error", "Structure error")

        assert len(report.get_by_severity("info")) == 1
        assert len(report.get_by_severity("warning")) == 1
        assert len(report.get_by_severity("error")) == 1

        assert len(report.get_by_category("style")) == 1
        assert len(report.get_by_category("optimization")) == 1

        assert report.has_issues  # Has warnings/errors

    def test_optimization_report_impact(self) -> None:
        """Test optimization report impact tracking."""
        report = OptimizationReport()

        report.add_suggestion("rule1", "type1", "Low impact", "low")
        report.add_suggestion("rule2", "type2", "Medium impact", "medium")
        report.add_suggestion("rule3", "type3", "High impact", "high")
        report.add_suggestion("rule4", "type3", "Another high", "high")

        assert report.high_impact_count == 2
        assert len(report.suggestions) == 4



from __future__ import annotations

import pytest

from yaraast.lexer.tokens import TokenType as T
from yaraast.yaral._shared import YaraLParserError
from yaraast.yaral.ast_nodes import (
    BinaryCondition,
    EventCountCondition,
    EventExistsCondition,
    UnaryCondition,
    VariableComparisonCondition,
)
from yaraast.yaral.lexer import YaraLToken
from yaraast.yaral.parser import YaraLParser
from yaraast.yaral.tokens import YaraLTokenType


def _tok(
    token_type: T, value: object, line: int = 1, yaral_type: YaraLTokenType | None = None
) -> YaraLToken:
    return YaraLToken(
        type=token_type,
        value=value,
        line=line,
        column=1,
        length=1,
        yaral_type=yaral_type,
    )


def _set_tokens(parser: YaraLParser, tokens: list[YaraLToken]) -> None:
    parser.tokens = tokens
    parser.current = 0


def test_parse_condition_section_and_boolean_precedence() -> None:
    parser = YaraLParser("")
    _set_tokens(
        parser,
        [
            _tok(T.IDENTIFIER, "condition"),
            _tok(T.COLON, ":"),
            _tok(T.IDENTIFIER, "not"),
            _tok(T.STRING_IDENTIFIER, "$e1", yaral_type=YaraLTokenType.EVENT_VAR),
            _tok(T.IDENTIFIER, "and"),
            _tok(T.STRING_IDENTIFIER, "$e2", yaral_type=YaraLTokenType.EVENT_VAR),
            _tok(T.IDENTIFIER, "or"),
            _tok(T.STRING_IDENTIFIER, "$e3", yaral_type=YaraLTokenType.EVENT_VAR),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )

    section = parser._parse_condition_section()
    assert isinstance(section.expression, BinaryCondition)
    assert isinstance(section.expression.left, BinaryCondition)
    assert isinstance(section.expression.left.left, UnaryCondition)


def test_parse_primary_condition_count_and_comparison_variants() -> None:
    parser = YaraLParser("")
    _set_tokens(
        parser,
        [
            _tok(T.STRING_COUNT, "#"),
            _tok(T.IDENTIFIER, "e"),
            _tok(T.EQ, "=="),
            _tok(T.INTEGER, "5"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )
    cond = parser._parse_primary_condition()
    assert isinstance(cond, EventCountCondition)
    assert cond.operator == "=="

    parser2 = YaraLParser("")
    _set_tokens(
        parser2,
        [
            _tok(T.STRING_IDENTIFIER, "$v", yaral_type=YaraLTokenType.EVENT_VAR),
            _tok(T.GE, ">="),
            _tok(T.IDENTIFIER, "other"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )
    cmp_cond = parser2._parse_primary_condition()
    assert isinstance(cmp_cond, VariableComparisonCondition)
    assert cmp_cond.operator == ">="

    parser3 = YaraLParser("")
    _set_tokens(
        parser3,
        [
            _tok(T.STRING_IDENTIFIER, "$exists", yaral_type=YaraLTokenType.EVENT_VAR),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )
    exists = parser3._parse_primary_condition()
    assert isinstance(exists, EventExistsCondition)
    assert exists.event == "exists"


def test_parse_primary_condition_identifier_fallback_and_errors() -> None:
    parser = YaraLParser("")
    _set_tokens(
        parser,
        [
            _tok(T.IDENTIFIER, "ev"),
            _tok(T.NEQ, "!="),
            _tok(T.STRING, "x"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )
    cond = parser._parse_primary_condition()
    assert isinstance(cond, VariableComparisonCondition)
    assert cond.operator == "!="

    parser2 = YaraLParser("")
    _set_tokens(
        parser2, [_tok(T.IDENTIFIER, "only_name"), _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF)]
    )
    exists = parser2._parse_primary_condition()
    assert isinstance(exists, EventExistsCondition)
    assert exists.event == "only_name"

    parser3 = YaraLParser("")
    _set_tokens(
        parser3,
        [
            _tok(T.STRING_COUNT, "#"),
            _tok(T.IDENTIFIER, "e"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )
    with pytest.raises(YaraLParserError, match="Expected comparison operator"):
        parser3._parse_primary_condition()

    parser4 = YaraLParser("")
    _set_tokens(
        parser4,
        [
            _tok(T.STRING_IDENTIFIER, "$e", yaral_type=YaraLTokenType.EVENT_VAR),
            _tok(T.LT, "<"),
            _tok(T.RPAREN, ")"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )
    with pytest.raises(YaraLParserError, match="Expected value after comparison operator"):
        parser4._parse_primary_condition()

    parser5 = YaraLParser("")
    _set_tokens(parser5, [_tok(T.RBRACE, "}"), _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF)])
    with pytest.raises(YaraLParserError, match="Unexpected token in condition"):
        parser5._parse_primary_condition()


@pytest.mark.parametrize(
    ("operator_token", "expected_operator"),
    [
        (T.NEQ, "!="),
        (T.LE, "<="),
        (T.EQ, "=="),
    ],
)
def test_parse_primary_condition_count_comparison_operator_variants(
    operator_token: T,
    expected_operator: str,
) -> None:
    parser = YaraLParser("")
    _set_tokens(
        parser,
        [
            _tok(T.STRING_COUNT, "#"),
            _tok(T.IDENTIFIER, "evt"),
            _tok(operator_token, expected_operator),
            _tok(T.INTEGER, "2"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )

    cond = parser._parse_primary_condition()
    assert isinstance(cond, EventCountCondition)
    assert cond.operator == expected_operator


@pytest.mark.parametrize(
    ("operator_token", "expected_operator", "value_token", "value", "yaral_type"),
    [
        (T.EQ, "==", T.STRING, "abc", None),
        (T.NEQ, "!=", T.STRING_IDENTIFIER, "$other", None),
        (T.LE, "<=", T.IDENTIFIER, "field_name", None),
        (T.NEQ, "!=", T.STRING_IDENTIFIER, "$peer", YaraLTokenType.EVENT_VAR),
    ],
)
def test_parse_primary_condition_variable_comparison_variants(
    operator_token: T,
    expected_operator: str,
    value_token: T,
    value: object,
    yaral_type: YaraLTokenType | None,
) -> None:
    parser = YaraLParser("")
    _set_tokens(
        parser,
        [
            _tok(T.STRING_IDENTIFIER, "$value", yaral_type=YaraLTokenType.EVENT_VAR),
            _tok(operator_token, expected_operator),
            _tok(value_token, value, yaral_type=yaral_type),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )

    cond = parser._parse_primary_condition()
    assert isinstance(cond, VariableComparisonCondition)
    assert cond.operator == expected_operator
    assert cond.value == value


@pytest.mark.parametrize(
    ("operator_token", "expected_operator", "value_token", "value", "yaral_type"),
    [
        (T.LT, "<", T.INTEGER, "7", None),
        (T.GE, ">=", T.STRING, "x", None),
        (T.LE, "<=", T.IDENTIFIER, "user", None),
        (T.EQ, "==", T.STRING_IDENTIFIER, "$peer", YaraLTokenType.EVENT_VAR),
        (T.NEQ, "!=", T.STRING_IDENTIFIER, "$peer", None),
    ],
)
def test_parse_primary_condition_identifier_fallback_operator_variants(
    operator_token: T,
    expected_operator: str,
    value_token: T,
    value: object,
    yaral_type: YaraLTokenType | None,
) -> None:
    parser = YaraLParser("")
    _set_tokens(
        parser,
        [
            _tok(T.IDENTIFIER, "field"),
            _tok(operator_token, expected_operator),
            _tok(value_token, value, yaral_type=yaral_type),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )

    cond = parser._parse_primary_condition()
    assert isinstance(cond, VariableComparisonCondition)
    assert cond.operator == expected_operator


def test_parse_primary_condition_identifier_missing_value_error() -> None:
    parser = YaraLParser("")
    _set_tokens(
        parser,
        [
            _tok(T.IDENTIFIER, "field"),
            _tok(T.NEQ, "!="),
            _tok(T.RPAREN, ")"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )

    with pytest.raises(YaraLParserError, match="Expected value after comparison operator"):
        parser._parse_primary_condition()


def test_parse_primary_condition_explicit_neq_branches_for_var_and_identifier() -> None:
    parser = YaraLParser("")
    _set_tokens(
        parser,
        [
            _tok(T.STRING_IDENTIFIER, "$field"),
            _tok(T.NEQ, "!="),
            _tok(T.INTEGER, "9"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )
    cond = parser._parse_primary_condition()
    assert isinstance(cond, VariableComparisonCondition)
    assert cond.variable == "$field"
    assert cond.operator == "!="
    assert cond.value == 9

    parser2 = YaraLParser("")
    _set_tokens(
        parser2,
        [
            _tok(T.IDENTIFIER, "field"),
            _tok(T.NEQ, "!="),
            _tok(T.INTEGER, "11"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )
    cond2 = parser2._parse_primary_condition()
    assert isinstance(cond2, VariableComparisonCondition)
    assert cond2.variable == "field"
    assert cond2.operator == "!="
    assert cond2.value == 11


def test_parse_primary_condition_parenthesized_expression() -> None:
    parser = YaraLParser("")
    _set_tokens(
        parser,
        [
            _tok(T.LPAREN, "("),
            _tok(T.STRING_IDENTIFIER, "$e1", yaral_type=YaraLTokenType.EVENT_VAR),
            _tok(T.IDENTIFIER, "and"),
            _tok(T.STRING_IDENTIFIER, "$e2", yaral_type=YaraLTokenType.EVENT_VAR),
            _tok(T.RPAREN, ")"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )

    cond = parser._parse_primary_condition()
    assert isinstance(cond, BinaryCondition)


@pytest.mark.parametrize(
    ("operator_token", "expected_operator"),
    [
        (T.GT, ">"),
        (T.LT, "<"),
        (T.GE, ">="),
    ],
)
def test_parse_primary_condition_count_remaining_operator_variants(
    operator_token: T,
    expected_operator: str,
) -> None:
    parser = YaraLParser("")
    _set_tokens(
        parser,
        [
            _tok(T.STRING_COUNT, "#"),
            _tok(T.IDENTIFIER, "evt"),
            _tok(operator_token, expected_operator),
            _tok(T.INTEGER, "9"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )

    cond = parser._parse_primary_condition()
    assert isinstance(cond, EventCountCondition)
    assert cond.operator == expected_operator


def test_parse_primary_condition_variable_gt_integer_and_fallback_gt() -> None:
    parser = YaraLParser("")
    _set_tokens(
        parser,
        [
            _tok(T.STRING_IDENTIFIER, "$metric", yaral_type=YaraLTokenType.EVENT_VAR),
            _tok(T.GT, ">"),
            _tok(T.INTEGER, "3"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )
    cond = parser._parse_primary_condition()
    assert isinstance(cond, VariableComparisonCondition)
    assert cond.operator == ">"
    assert cond.value == 3

    parser2 = YaraLParser("")
    _set_tokens(
        parser2,
        [
            _tok(T.IDENTIFIER, "count"),
            _tok(T.GT, ">"),
            _tok(T.INTEGER, "1"),
            _tok(T.EOF, None, yaral_type=YaraLTokenType.EOF),
        ],
    )
    cond2 = parser2._parse_primary_condition()
    assert isinstance(cond2, VariableComparisonCondition)
    assert cond2.operator == ">"
    assert cond2.value == 1
